"""Scripts for testing connection.py.

Author(s): Erwin de Gelder
"""

from pathlib import Path

import matplotlib as mpl
import matplotlib.pyplot as plt
import pytest
from matplotlib.axes import Axes

from traffic_scene_renderer import (
    Connection,
    Crossing,
    InvalidConnectionError,
    Vertex,
    Way,
    WayOptions,
)

from .test_static_objects import save_fig
from .test_way import plot_way

mpl.use("Agg")


def plot_connection(axes: Axes, *connections: Connection) -> None:
    for connection in connections:
        connection.process()
        if connection.crossing is not None:
            connection.crossing.plot(axes)
    ways = set()
    for connection in connections:
        for way in (connection.way1, connection.way2):
            if way not in ways:
                way.process()
                plot_way(axes, way)
                ways.add(way)


def test_simple_connection() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-5, 5)
    vertices = [
        Vertex(0, -10, 0),
        Vertex(1, -5, 0),
        Vertex(2, 0, 0),
        Vertex(3, 5, 0),
        Vertex(4, 10, 0),
    ]
    way1 = Way(vertices[:3], WayOptions(side_color=(1, 0, 0), line_color=(1, 0, 0)))
    way2 = Way(vertices[2:], WayOptions(side_color=(0, 0, 1), line_color=(0, 0, 1)))
    connection = Connection(0, way1, way2)
    plot_connection(axes, connection)
    save_fig(fig, axes, Path("connection") / "simple_connection.png", 10)


def test_lane_addition() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-8, 8)
    vertices = [
        Vertex(0, -10, 0),
        Vertex(1, -5, 0),
        Vertex(2, 0, 0),
        Vertex(3, 5, -0.7),
        Vertex(4, 10, -1.4),
    ]
    way1 = Way(vertices[2::-1], WayOptions(side_color=(1, 0, 0), line_color=(1, 0, 0)))
    way2 = Way(vertices[:1:-1], WayOptions(side_color=(0, 0, 1), line_color=(0, 0, 1), nlanes=3))
    connection = Connection(0, way2, way1)
    plot_connection(axes, connection)
    save_fig(fig, axes, Path("connection") / "lane_addition.png", 10)


def test_lane_merge() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-8, 8)
    vertices = [
        Vertex(0, -10, 0),
        Vertex(1, -5, 0),
        Vertex(2, 0, 0),
        Vertex(3, 5, 0.7),
        Vertex(4, 10, 1.4),
    ]
    way1 = Way(vertices[2::-1], WayOptions(side_color=(1, 0, 0), line_color=(1, 0, 0)))
    way2 = Way(
        vertices[2:],
        WayOptions(
            side_color=(0, 0, 1),
            line_color=(0, 0, 1),
            nlanes=3,
            turnlanes="left|none|none",
            marker_border_color=(0, 0, 0.5),
        ),
    )
    connection = Connection(0, way1, way2)
    plot_connection(axes, connection)
    way2.plot_markers(axes)
    save_fig(fig, axes, Path("connection") / "lane_merge.png", 10)


def test_connection_as_crossing() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-5, 10)
    vertices = [Vertex(0, -10, 0), Vertex(1, 0, 0), Vertex(2, 2, 12)]
    way1 = Way([vertices[0], vertices[1]], WayOptions(side_color=(1, 0, 0), line_color=(1, 0, 0)))
    way2 = Way([vertices[2], vertices[1]], WayOptions(side_color=(0, 0, 1), line_color=(0, 0, 1)))
    connection = Connection(0, way1, way2)
    plot_connection(axes, connection)
    save_fig(fig, axes, Path("connection") / "connection_as_crossing.png", 10)


def test_invalid_connection_error() -> None:
    way1 = Way([Vertex(0, -10, 0), Vertex(1, 0, 0)])
    way2 = Way([Vertex(2, -10, 10), Vertex(3, 0, 10)])
    try:
        Connection(0, way1, way2)
    except InvalidConnectionError:
        pass
    else:
        pytest.fail("InvalidConnectionError should have been raised.")


def test_check_crossing_reprocessing() -> None:
    vertices = [
        Vertex(0, -20, 0),
        Vertex(1, -10, 0),
        Vertex(2, -10, 10),
        Vertex(3, 0, 0),
        Vertex(4, 10, 0),
        Vertex(5, 10, 10),
        Vertex(6, 20, 0),
    ]
    ways = [
        Way([vertices[0], vertices[1]]),
        Way([vertices[1], vertices[2]]),
        Way([vertices[1], vertices[3]]),
        Way([vertices[3], vertices[4]]),
        Way([vertices[4], vertices[5]]),
        Way([vertices[4], vertices[6]]),
    ]
    crossing1 = Crossing(10, vertices[1], ways[:3])
    crossing2 = Crossing(20, vertices[4], ways[3:])
    connection = Connection(0, ways[2], ways[3])
    assert connection.compute_distance_to_next_vertex()[1] == [crossing1.idx, crossing2.idx]



"""Constructing crossings.

Author(s): Erwin de Gelder
"""

import warnings
from typing import List, NamedTuple, Optional, Tuple, Union

import numpy as np
from matplotlib import patches
from matplotlib.axes import Axes

from .options import Options
from .vertex import Vertex
from .way import Way, XYData


class CornerInfo(NamedTuple):
    """Tuple for storing information on a single corner."""

    radius: float
    circle_x: float
    circle_y: float
    lambda_left: float
    lambda_right: float
    i_way2: int


class StartDirection(NamedTuple):
    """Tuple for storing information of a single arm of the crossing."""

    x_init: float
    y_init: float
    x_dir: float
    y_dir: float


class IntersectionWays(NamedTuple):
    """Tuple for storing information of two adjacent arms of a crossing."""

    x: float
    y: float
    lambda_left: float
    lambda_right: float
    way1: StartDirection
    way2: StartDirection


class CrossingOptions(Options):
    """Object containing the objects for a crossing."""

    color: Optional[Tuple[float, float, float]] = None
    edge_color: Tuple[float, float, float] = (0, 0, 0)
    n_corner_pieces: int = 20
    radius: float = -1
    zebra_color: Tuple[float, float, float] = (1, 1, 1)
    zebra_width: float = 0.8
    zebra_square_markings: bool = False
    zebra_square_markings_color: Tuple[float, float, float] = (235 / 255, 203 / 255, 108 / 255)
    zorder: int = 0
    roundabout: bool = False
    roundabout_outer_radius: float = 6
    roundabout_inner_radius: float = 2
    roundabout_inner_color: Tuple[float, float, float] = (0.8, 1, 0.8)
    default_radius: float = 2.0
    default_radius_footway: float = 0.01


class CrossingCircles(Options):
    """Parameters of a crossing concerning the circles between two ways."""

    x_data: np.ndarray = np.array([])
    y_data: np.ndarray = np.array([])
    radii: np.ndarray = np.array([])


class CrossingWayParameters(Options):
    """Parameters of a crossing concerning the ways that are connected to it."""

    at_start: np.ndarray = np.array([], dtype=bool)
    lambda_left: np.ndarray = np.array([])
    lambda_right: np.ndarray = np.array([])
    angles: np.ndarray = np.array([])


class CrossingParameters(Options):
    """Object that contains various parameters of a crossing."""

    zebra: bool = False
    processed: bool = False
    part_of_big_crossing: bool = False

    def __init__(
        self, **kwargs: Union[bool, CrossingCircles, XYData, CrossingWayParameters]
    ) -> None:
        """Initialize object that contains various parameters of a crossing.

        :param kwargs: Any parameters that are to be set can be passed via kwargs.
        """
        self.circles = CrossingCircles()
        self.plot = XYData()
        self.way = CrossingWayParameters()
        Options.__init__(self, **kwargs)


class Crossing:
    """Constructing a crossing with multiple ways.

    This object is used for drawing a crossing between two or more ways.

    Attributes:
        idx (int): The index of the crossing.
        options (CrossingOptions): Options that change the way the crossing is rendered.
        vertex (Vertex): The vertex that is shared among the different ways.
        ways (List[Way]): The ways that are connected to the crossing.
        parms (CrossingParms): Parameters of the crossing that are used for rendering.
    """

    def __init__(
        self, index: int, vertex: Vertex, ways: List[Way], options: Optional[CrossingOptions] = None
    ) -> None:
        """Initialize a crossing.

        :param index: Nonnegative index of the crossing.
        :param vertex: The vertex at which the crossing is located.
        :param ways: The ways that are connected to the crossing.
        :param options: Any options of the crossing, see CrossingOptions.
        """
        self.options = CrossingOptions() if options is None else options
        self.idx = index
        self.vertex = vertex
        self.ways = ways

        self.parms = CrossingParameters(
            zebra=np.sum([way.options.highway == "footway" for way in self.ways]) == 2,  # noqa: PLR2004
            way=CrossingWayParameters(
                at_start=np.zeros(len(self.ways), dtype=bool),
                angles=np.array([np.arctan2(*self.direction(way)) for way in self.ways]),
                lambda_left=np.zeros(len(self.ways)),
                lambda_right=np.zeros(len(self.ways)),
            ),
            circles=CrossingCircles(
                radii=np.ones(len(self.ways)) * -1,
                x_data=np.zeros(len(self.ways)),
                y_data=np.zeros(len(self.ways)),
            ),
            plot=XYData(
                x_data=np.zeros((self.options.n_corner_pieces + 2, len(self.ways))),
                y_data=np.zeros((self.options.n_corner_pieces + 2, len(self.ways))),
            ),
        )
        self._determine_color()

        # Reorder the ways, such that they are in clockwise order
        iways = np.argsort(self.parms.way.angles)  # type: np.ndarray
        self.ways = [self.ways[i] for i in iways]
        self.parms.way.angles = np.sort(self.parms.way.angles)

        # Determine if either the start or end of a way is connected to this crossing.
        for i_way, way in enumerate(self.ways):
            if way.ivs[0] == self.vertex.idx:  # Only check if at start, otherwise assume at end.
                self.parms.way.at_start[i_way] = True
                way.parms.crossing.i_start = self.idx
            else:
                way.parms.crossing.i_end = self.idx

        # Set reference to crossing correct for the ways
        # We could not do this earlier, because ways are potentially reordered
        for i_way, way in enumerate(self.ways):
            if self.parms.way.at_start[i_way]:
                way.parms.crossing.i_start = self.idx
            else:
                way.parms.crossing.i_end = self.idx

    def compute_corner_info(self, i_way1: int) -> CornerInfo:
        """Compute information of the corner.

        The returned CornerInfo contains the radius of the corner, (x,y)-coordinates
        of the circle, and the part of the way that is used by the corner (lambda_left
        and  lambda_right) as a ratio of the segment of the way that is connected to
        the crossing.

        :param i_way1: Index of first way.
        :return: A CornerInfo structure with the information.
        """
        i_way2 = i_way1 + 1 if i_way1 < len(self.ways) - 1 else 0  # Index of second way.
        angle = np.mod(self.parms.way.angles[i_way2] - self.parms.way.angles[i_way1], 2 * np.pi)
        if angle >= np.pi:
            # In this case, the radius equals the half width of the roads (assumed to be similar).
            # The center of the circle is simply the coordinate of the vertex.
            # The end of the way is used, so lambda=0.
            return CornerInfo(
                radius=self.ways[i_way1].parms.hwidth,
                circle_x=self.vertex.xcoordinate,
                circle_y=self.vertex.ycoordinate,
                lambda_left=0,
                lambda_right=0,
                i_way2=i_way2,
            )
        # Angle is less than 180 degrees
        # For the center of the corner cicle, we need to determine stuff of lines that go
        # through center.
        radius = (
            self.options.radius
            if self.parms.circles.radii[i_way1] < 0
            else self.parms.circles.radii[i_way1]
        )
        intersection = self.compute_intersection(
            i_way1,
            i_way2,
            self.ways[i_way1].parms.hwidth + radius,
            self.ways[i_way2].parms.hwidth + radius,
        )

        # Check if lambda is not larger than 1. If so, we need to decrease radius
        max_right = 0.45 if len(self.ways[i_way1].vertices) == 2 else 0.9  # noqa: PLR2004
        max_left = 0.45 if len(self.ways[i_way2].vertices) == 2 else 0.9  # noqa: PLR2004
        if intersection.lambda_right > max_right or intersection.lambda_left > max_left:
            # Compute intersection of road boundaries
            roadside = self.compute_intersection(
                i_way1, i_way2, self.ways[i_way1].parms.hwidth, self.ways[i_way2].parms.hwidth
            )
            dist = min(
                np.hypot(
                    roadside.way1.x_init + max_right * roadside.way1.x_dir - roadside.x,
                    roadside.way1.y_init + max_right * roadside.way1.y_dir - roadside.y,
                ),
                np.hypot(
                    roadside.way2.x_init + max_left * roadside.way2.x_dir - roadside.x,
                    roadside.way2.y_init + max_left * roadside.way2.y_dir - roadside.y,
                ),
            )
            radius = dist * np.tan(angle / 2)

            # Compute again the center of the circle
            intersection = self.compute_intersection(
                i_way1,
                i_way2,
                self.ways[i_way1].parms.hwidth + radius,
                self.ways[i_way2].parms.hwidth + radius,
            )
        return CornerInfo(
            radius=radius,
            circle_x=intersection.x,
            circle_y=intersection.y,
            i_way2=i_way2,
            lambda_left=intersection.lambda_left,
            lambda_right=intersection.lambda_right,
        )

    def process(self) -> None:
        """Process the crossing, i.e., create the boundaries."""
        self.options.radius = self.compute_radius()
        if self.options.roundabout:
            self.process_roundabout()
            return

        # Loop through all corners to compute the boundary location of the circle
        nways = len(self.ways)
        for i in range(nways):  # Index of first way
            corner_info = self.compute_corner_info(i)
            self.parms.circles.radii[i] = corner_info.radius
            self.parms.circles.x_data[i] = corner_info.circle_x
            self.parms.circles.y_data[i] = corner_info.circle_y
            self.parms.way.lambda_left[corner_info.i_way2] = corner_info.lambda_left
            self.parms.way.lambda_right[i] = corner_info.lambda_right

        # Loop again through all corners to compute the coordinates of the boundary of the road.
        for i in range(nways):  # Index of first way
            # Determine (x,y) coordinates of the corners
            i_way2 = i + 1 if i < nways - 1 else 0  # Index of second way
            angle = np.mod(self.parms.way.angles[i_way2] - self.parms.way.angles[i], 2 * np.pi)
            if angle >= np.pi:
                theta0 = self.parms.way.angles[i] + np.pi / 2
            else:  # Angle is less than 180 degrees
                theta0 = self.parms.way.angles[i] - np.pi / 2
            theta1 = theta0 - np.pi + angle
            theta = np.linspace(theta0, theta1, self.options.n_corner_pieces)
            self.parms.plot.x_data[1:-1, i] = (
                self.parms.circles.x_data[i] + np.sin(theta) * self.parms.circles.radii[i]
            )
            self.parms.plot.y_data[1:-1, i] = (
                self.parms.circles.y_data[i] + np.cos(theta) * self.parms.circles.radii[i]
            )

            # Determine the (x,y) coordinates of the end of the crossing.
            way1 = self.start_and_direction(i, self.ways[i].parms.hwidth, left=False)
            way2 = self.start_and_direction(i_way2, self.ways[i_way2].parms.hwidth, left=True)
            self.parms.plot.x_data[0, i] = way1.x_init + way1.x_dir * max(
                self.parms.way.lambda_left[i], self.parms.way.lambda_right[i]
            )
            self.parms.plot.y_data[0, i] = way1.y_init + way1.y_dir * max(
                self.parms.way.lambda_left[i], self.parms.way.lambda_right[i]
            )
            self.parms.plot.x_data[-1, i] = way2.x_init + way2.x_dir * max(
                self.parms.way.lambda_left[i_way2], self.parms.way.lambda_right[i_way2]
            )
            self.parms.plot.y_data[-1, i] = way2.y_init + way2.y_dir * max(
                self.parms.way.lambda_left[i_way2], self.parms.way.lambda_right[i_way2]
            )

        # Set the lambdas for the individual ways.
        for i, way in enumerate(self.ways):
            if self.parms.way.at_start[i]:
                way.parms.crossing.start_lambda_left = self.parms.way.lambda_left[i]
                way.parms.crossing.start_lambda_right = self.parms.way.lambda_right[i]
            else:
                way.parms.crossing.end_lambda_left = self.parms.way.lambda_left[i]
                way.parms.crossing.end_lambda_right = self.parms.way.lambda_right[i]

        # Set flag that this crossing is processed.
        self.parms.processed = True

    def compute_corner_info_roundabout(self, i_way: int, *, left: bool) -> CornerInfo:
        """Compute the location of the corner.

        :param i_way: The index of the way.
        :param left: Whether the corner is on the left or the right.
        :return: Information about the corner.
        """
        radius = self.options.radius
        way = self.start_and_direction(
            i_way, offset=self.ways[i_way].parms.hwidth + radius, left=left
        )
        square = way.x_dir**2 + way.y_dir**2
        linear = 2 * (
            (way.x_init - self.vertex.xcoordinate) * way.x_dir
            + (way.y_init - self.vertex.ycoordinate) * way.y_dir
        )
        constant = (
            (way.x_init - self.vertex.xcoordinate) ** 2
            + (way.y_init - self.vertex.ycoordinate) ** 2
            - (self.options.roundabout_outer_radius + radius) ** 2
        )
        lam = (-linear + np.sqrt(linear**2 - 4 * square * constant)) / (2 * square)
        return CornerInfo(
            radius=radius,
            lambda_right=lam,
            lambda_left=lam,
            i_way2=0,
            circle_x=way.x_init + way.x_dir * lam,
            circle_y=way.y_init + way.y_dir * lam,
        )

    def process_roundabout(self) -> None:
        """Process the roundabout, i.e., create the boundaries."""
        # For each corner, we will have three circles.
        self.parms.circles.radii = np.ones(len(self.ways) * 3) * -1
        self.parms.circles.y_data = np.zeros_like(self.parms.circles.radii)
        self.parms.circles.x_data = np.zeros_like(self.parms.circles.radii)

        # The plot data also includes the inner circle (last row).
        self.parms.plot.x_data = np.zeros((self.options.n_corner_pieces * 3 + 2, len(self.ways)))
        self.parms.plot.y_data = np.zeros_like(self.parms.plot.x_data)

        # Loop through all ways and compute the centers of the circles.
        nways = len(self.ways)
        for i in range(nways):
            for j, left in enumerate([False, True]):
                corner = self.compute_corner_info_roundabout(i, left=left)
                self.parms.circles.radii[i * 3 - j] = corner.radius
                self.parms.circles.x_data[i * 3 - j] = corner.circle_x
                self.parms.circles.y_data[i * 3 - j] = corner.circle_y
                if j == 0:  # This has to be done only once...
                    self.parms.way.lambda_right[i] = corner.lambda_right
                    self.parms.way.lambda_left[i] = corner.lambda_left  # Is the same.
            self.parms.circles.radii[i * 3 + 1] = self.options.roundabout_outer_radius
            self.parms.circles.x_data[i * 3 + 1] = self.vertex.xcoordinate
            self.parms.circles.y_data[i * 3 + 1] = self.vertex.ycoordinate

        # Define the borders
        for i in range(nways):
            i_way2 = i + 1 if i < nways - 1 else 0  # Index of second way

            # Determine the (x,y) coordinates of the end of the crossing.
            way1 = self.start_and_direction(i, self.ways[i].parms.hwidth, left=False)
            way2 = self.start_and_direction(i_way2, self.ways[i_way2].parms.hwidth, left=True)
            self.parms.plot.x_data[0, i] = way1.x_init + way1.x_dir * max(
                self.parms.way.lambda_left[i], self.parms.way.lambda_right[i]
            )
            self.parms.plot.y_data[0, i] = way1.y_init + way1.y_dir * max(
                self.parms.way.lambda_left[i], self.parms.way.lambda_right[i]
            )
            self.parms.plot.x_data[-1, i] = way2.x_init + way2.x_dir * max(
                self.parms.way.lambda_left[i_way2], self.parms.way.lambda_right[i_way2]
            )
            self.parms.plot.y_data[-1, i] = way2.y_init + way2.y_dir * max(
                self.parms.way.lambda_left[i_way2], self.parms.way.lambda_right[i_way2]
            )

            # Compute corners
            x_data, y_data = self.roundabout_corner(i, i_way2, big_angle=False)
            self.parms.plot.x_data[1:-1, i] = x_data
            self.parms.plot.y_data[1:-1, i] = y_data

        # Set the lambdas for the individual ways.
        for i, way in enumerate(self.ways):
            if self.parms.way.at_start[i]:
                way.parms.crossing.start_lambda_left = self.parms.way.lambda_left[i]
                way.parms.crossing.start_lambda_right = self.parms.way.lambda_right[i]
            else:
                way.parms.crossing.end_lambda_left = self.parms.way.lambda_left[i]
                way.parms.crossing.end_lambda_right = self.parms.way.lambda_right[i]

        # Set flag that this crossing is processed.
        self.parms.processed = True

    def roundabout_corner(
        self, i_way1: int, i_way2: int, *, big_angle: bool, radius: Optional[float] = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Create a corner of a roundabout.

        The function is used to create the corners of a roundabout. However,
        the function can also be used to create the trajectory of a vehicle
        driving on the roundabout. If the vehicle is supposed to drive more than
        50% of the roundabout (i.e., > pi), big_angle should be set to True.

        :param i_way1: Index of the way at which the line starts.
        :param i_way2: Index of the way at which the line ends.
        :param big_angle: Whether the angle is larger than pi.
        :param radius: The radius of the circle.
        :return: (x,y) data of the line.
        """
        if radius is None:
            radius = self.options.roundabout_outer_radius
        begin_radius = self.options.radius + self.options.roundabout_outer_radius - radius
        way1 = self.start_and_direction(i_way1, 0, left=True)
        way2 = self.start_and_direction(i_way2, 0, left=True)
        x_data = np.zeros(3 * self.options.n_corner_pieces)
        y_data = np.zeros(3 * self.options.n_corner_pieces)
        theta0 = np.arctan2(-way1.y_dir, way1.x_dir)
        theta1 = np.arctan2(
            self.vertex.xcoordinate - self.parms.circles.x_data[i_way1 * 3],
            self.vertex.ycoordinate - self.parms.circles.y_data[i_way1 * 3],
        )
        theta0 = correct_angle(theta0, theta1, big_angle=big_angle)
        theta = np.linspace(theta0, theta1, self.options.n_corner_pieces)
        x_data[: self.options.n_corner_pieces] = (
            self.parms.circles.x_data[i_way1 * 3] + np.sin(theta) * begin_radius
        )
        y_data[: self.options.n_corner_pieces] = (
            self.parms.circles.y_data[i_way1 * 3] + np.cos(theta) * begin_radius
        )
        theta0 = np.mod(theta1, 2 * np.pi) - np.pi
        theta1 = np.arctan2(
            self.parms.circles.x_data[i_way2 * 3 - 1] - self.vertex.xcoordinate,
            self.parms.circles.y_data[i_way2 * 3 - 1] - self.vertex.ycoordinate,
        )
        theta0 = correct_angle(theta0, theta1, big_angle=big_angle)
        theta = np.linspace(theta0, theta1, self.options.n_corner_pieces)
        x_data[self.options.n_corner_pieces : 2 * self.options.n_corner_pieces] = (
            self.vertex.xcoordinate + np.sin(theta) * radius
        )
        y_data[self.options.n_corner_pieces : 2 * self.options.n_corner_pieces] = (
            self.vertex.ycoordinate + np.cos(theta) * radius
        )
        theta0 = np.mod(theta1, 2 * np.pi) - np.pi
        theta1 = np.arctan2(way2.y_dir, -way2.x_dir)
        theta0 = correct_angle(theta0, theta1, big_angle=big_angle)
        theta = np.linspace(theta0, theta1, self.options.n_corner_pieces)
        x_data[2 * self.options.n_corner_pieces :] = (
            self.parms.circles.x_data[i_way2 * 3 - 1] + np.sin(theta) * begin_radius
        )
        y_data[2 * self.options.n_corner_pieces :] = (
            self.parms.circles.y_data[i_way2 * 3 - 1] + np.cos(theta) * begin_radius
        )
        return x_data, y_data

    def plot(self, axes: Axes) -> None:
        """Plot the crossing on an axes.

        :param axes: The axes that is used for plotting.
        """
        x_plot = np.reshape(self.parms.plot.x_data.T, (np.prod(self.parms.plot.x_data.shape), 1))
        y_plot = np.reshape(self.parms.plot.y_data.T, (np.prod(self.parms.plot.y_data.shape), 1))
        axes.add_patch(
            patches.Polygon(
                np.concatenate((x_plot, y_plot), axis=1),
                color=self.options.color,
                zorder=self.options.zorder,
            )
        )
        if self.options.edge_color is not None:
            axes.plot(
                self.parms.plot.x_data,
                self.parms.plot.y_data,
                color=self.options.edge_color,
                zorder=self.options.zorder,
            )

        if self.options.roundabout:
            theta = np.linspace(0, 2 * np.pi, 4 * self.options.n_corner_pieces)
            x_plot = np.sin(theta) * self.options.roundabout_inner_radius + self.vertex.xcoordinate
            y_plot = np.cos(theta) * self.options.roundabout_inner_radius + self.vertex.ycoordinate
            axes.add_patch(
                patches.Polygon(
                    np.concatenate(([x_plot], [y_plot]), axis=0).T,
                    color=self.options.roundabout_inner_color,
                    zorder=self.options.zorder,
                )
            )
            axes.plot(x_plot, y_plot, color=self.options.edge_color, zorder=self.options.zorder)

        if self.parms.zebra:
            self.plot_zebra(axes)

    def plot_zebra(self, axes: Axes) -> None:
        """Plot the zebra strips.

        :param axes: The axes that is used for plotting.
        """
        # Step 1: Find the endpoints of the zebra crossing
        x_footway, y_footway, x_road, y_road = self.find_endpoints_zebra()
        if len(x_footway) != 2:  # noqa: PLR2004
            warnings.warn(
                "No zebra crossing plotted, because not 2 footways connected to crossing",
                stacklevel=2,
            )
            return
        if len(x_road) != 2:  # noqa: PLR2004
            warnings.warn(
                "No zebra crossing plotted, because not 2 roadways connected to crossing",
                stacklevel=2,
            )
            return

        # Step 2: determine coordinates of a stripe with center (0, 0).
        nstripes, dxf, dyf, xstripe, ystripe = self.compute_coordinates_zebra(
            x_footway, y_footway, x_road, y_road
        )

        # Step 3: Draw the stripes.
        for i in range(nstripes):
            x_center = (0.5 + i) / nstripes * dxf + x_footway[0]
            y_center = (0.5 + i) / nstripes * dyf + y_footway[0]
            axes.fill(x_center + xstripe, y_center + ystripe, color=self.options.zebra_color)

        # Add yellow squares
        if self.options.zebra_square_markings:
            footpath_width = np.hypot(x_road[1] - x_road[0], y_road[1] - y_road[0])
            road_width = np.hypot(dxf, dyf)
            size_square = road_width / nstripes / 2
            x_square = (
                np.array([dyf - dxf, -dyf - dxf, -dyf + dxf, dyf + dxf])
                / road_width
                * size_square
                / 2
            )
            y_square = (
                np.array([-dyf - dxf, -dyf + dxf, dyf + dxf, dyf - dxf])
                / road_width
                * size_square
                / 2
            )
            for i in range(nstripes):
                x_center = (0.5 + i) / nstripes * dxf + x_footway[0]
                y_center = (0.5 + i) / nstripes * dyf + y_footway[0]
                for direction in [1, -1]:
                    xc_square = (
                        x_center
                        + direction / road_width * (footpath_width / 2 + 2 * size_square) * dyf
                    )
                    yc_square = (
                        y_center
                        - direction / road_width * (footpath_width / 2 + 2 * size_square) * dxf
                    )
                    axes.fill(
                        x_square + xc_square,
                        y_square + yc_square,
                        color=self.options.zebra_square_markings_color,
                    )
                    if i:
                        xc_square -= (dyf * direction + dxf) / road_width * size_square
                        yc_square += (dxf * direction - dyf) / road_width * size_square
                        axes.fill(
                            x_square + xc_square,
                            y_square + yc_square,
                            color=self.options.zebra_square_markings_color,
                        )

    def find_endpoints_zebra(self) -> Tuple[List, List, List, List]:
        """Find the endpoints of the zebra crossing.

        :return: A list of the coordinates of the footway and a list of the coordinates of the
                 road.
        """
        x_footway, y_footway = [], []  # Footway
        x_road, y_road = [], []  # Road
        for i, way in enumerate(self.ways):
            max_lambda = max(self.parms.way.lambda_left[i], self.parms.way.lambda_right[i])
            start_dir = self.start_and_direction(i, 0, left=True)
            if way.options.highway == "footway":
                x_footway.append(start_dir.x_init + max_lambda * start_dir.x_dir)
                y_footway.append(start_dir.y_init + max_lambda * start_dir.y_dir)
            else:
                x_road.append(start_dir.x_init + max_lambda * start_dir.x_dir)
                y_road.append(start_dir.y_init + max_lambda * start_dir.y_dir)
        return x_footway, y_footway, x_road, y_road

    def compute_coordinates_zebra(
        self, x_footway: List, y_footway: List, x_road: List, y_road: List
    ) -> Tuple[int, float, float, np.ndarray, np.ndarray]:
        """Determine coordinates of a zebra stripe with center (0, 0).

        :param x_footway: The x-coordinates of the footway.
        :param y_footway: The y-coordinates of the footway.
        :param x_road: The x-coordinates of the road.
        :param y_road: The y-coordinates of the road.
        :return: The number of stripes, the x-direction of the footway, the y-direction of the
                 footway, the x-coordinates of a stripe with center (0, 0), and the y-coordinates
                 of a stripe with center (0, 0).
        """
        dxf, dyf = x_footway[1] - x_footway[0], y_footway[1] - y_footway[0]
        dxr, dyr = x_road[1] - x_road[0], y_road[1] - y_road[0]
        nstripes = np.round(np.hypot(dxf, dyf) / self.options.zebra_width / 2).astype(int) * 2
        width = np.hypot(dxf, dyf) / nstripes / 2
        ratio = width / np.hypot(dxr, dyr)
        xstripe = (
            np.array([dxr - dyr * ratio, dxr + dyr * ratio, -dxr + dyr * ratio, -dxr - dyr * ratio])
            / 2
        )
        ystripe = (
            np.array([dyr + dxr * ratio, dyr - dxr * ratio, -dyr - dxr * ratio, -dyr + dxr * ratio])
            / 2
        )
        return nstripes, dxf, dyf, xstripe, ystripe

    def direction(self, index_or_way: Union[int, Way]) -> Tuple[float, float]:
        """Compute the direction (i.e., (dx, dy)) of the way.

        :param index_or_way: index of way or the way itself.
        :return: (x, y)-direction.
        """
        if isinstance(index_or_way, Way):
            way = index_or_way
            crossing_at_start = way.ivs[0] == self.vertex.idx
        else:
            way = self.ways[index_or_way]
            crossing_at_start = self.parms.way.at_start[index_or_way]
        if crossing_at_start:
            difference_x = way.vertices[1].xcoordinate - way.vertices[0].xcoordinate
            difference_y = way.vertices[1].ycoordinate - way.vertices[0].ycoordinate
        else:
            difference_x = way.vertices[-2].xcoordinate - way.vertices[-1].xcoordinate
            difference_y = way.vertices[-2].ycoordinate - way.vertices[-1].ycoordinate
        return difference_x, difference_y

    def start_and_direction(
        self, index_or_way: Union[int, Way], offset: float, *, left: bool
    ) -> StartDirection:
        """Compute starting position and direction of way i, given an offset.

        :param index_or_way: index of way or the way itself.
        :param offset: the total offset
        :param left: whether offset is on left or right side (seen when pointing away from the
                     crossing center)
        :return: StartDirection structure that contains starting point of way and its direction.
        """
        # Compute the direction first.
        x_dir, y_dir = self.direction(index_or_way)
        absdxy = np.hypot(x_dir, y_dir)

        # Compute offset starting position in case it is on the left side.
        xoffset = -y_dir / absdxy * offset
        yoffset = x_dir / absdxy * offset

        # Set initial position
        if left:
            x_init, y_init = self.vertex.xcoordinate + xoffset, self.vertex.ycoordinate + yoffset
        else:
            x_init, y_init = self.vertex.xcoordinate - xoffset, self.vertex.ycoordinate - yoffset

        return StartDirection(x_init=x_init, y_init=y_init, x_dir=x_dir, y_dir=y_dir)

    def compute_intersection(
        self, i_way1: int, i_way2: int, offset1: float, offset2: float
    ) -> IntersectionWays:
        """Compute the intersection of two ways given a certain offset.

        It returns the (x,y)-coordinate of the intersection, as well as the portions
        of the segments that are used of the two ways, i.e., lambda_left and
        lambda_right.

        :param i_way1: Index of first way.
        :param i_way2: Index of second way.
        :param offset1: The offset that is to be used for the first way.
        :param offset2: The offset that is to be used for the second way.
        :return: An IntersectionWays structure.
        """
        way1 = self.start_and_direction(i_way1, offset1, left=False)
        way2 = self.start_and_direction(i_way2, offset2, left=True)

        # The intersection is located where the lines meet, so compute this using linear algebra.
        matrix = np.array([[way1.x_dir, -way2.x_dir], [way1.y_dir, -way2.y_dir]])
        vector = np.array([way2.x_init - way1.x_init, way2.y_init - way1.y_init])
        solution = np.linalg.solve(matrix, vector)
        x_intersection = way1.x_init + solution[0] * way1.x_dir
        y_intersection = way1.y_init + solution[0] * way1.y_dir
        return IntersectionWays(
            x=x_intersection,
            y=y_intersection,
            way1=way1,
            way2=way2,
            lambda_right=solution[0],
            lambda_left=solution[1],
        )

    def compute_radius(self) -> float:
        """Compute the turning radius.

        :return: turning radius.
        """
        if self.options.radius != -1:
            return self.options.radius
        if self.parms.zebra:
            return 0.01
        if all(way.options.highway == "footway" for way in self.ways):
            return self.options.default_radius_footway
        return self.options.default_radius

    def _determine_color(self) -> None:
        """Determine the color based on the connected roads."""
        if not self.parms.zebra:
            # Just copy the color of the first way that is connected to the crossing.
            self.options.color = self.ways[0].options.color
            self.options.edge_color = self.ways[0].options.side_color
        else:
            # Copy the color of the first road that is not a footpath
            for way in self.ways:
                if way.options.highway != "footway":
                    self.options.color = way.options.color
                    self.options.edge_color = way.options.side_color


def correct_angle(angle1: float, angle2: float, *, big_angle: bool) -> float:
    """Correct the first angle if needed.

    The first angle is corrected with +2pi or -2pi, such that the difference
    between the two angles is less than pi.

    :param angle1: The first angle.
    :param angle2: The second angle.
    :param big_angle: If True, than the two angles should have a bigger
                      difference than pi.
    :return: The updated first angle.
    """
    if angle1 + np.pi < angle2 and not big_angle:
        return angle1 + 2 * np.pi
    if angle1 - np.pi > angle2 and not big_angle:
        return angle1 - 2 * np.pi
    if angle1 < angle2 < angle1 + np.pi and big_angle:
        return angle1 + 2 * np.pi
    if angle1 - np.pi < angle2 < angle1 and big_angle:
        return angle1 - 2 * np.pi
    return angle1



"""Plotting filled polygons like matplotlib, but provide additional option of having fixed color.

Author(s): Erwin de Gelder
"""

from typing import Any, Tuple

import numpy as np
from matplotlib import patches
from matplotlib.axes import Axes


class Polygon:
    """The polygon that is filled with the specified color.

    Attributes:
        xdata (np.ndarray): The x-coordinates of the vertices the polygon.
        ydata (np.ndarray): The y-coordinates of the vertices the polygon.
        patch (patches.Patch): Matplotlib's patch object that is used for plotting the polygon.
        fixed_color (bool): Whether or not the color is fixed.
    """

    def __init__(
        self,
        axes: Axes,
        xdata: np.ndarray,
        ydata: np.ndarray,
        *,
        fixed_color: bool = False,
        **kwargs: Any,  # noqa: ANN401  # Allow Any
    ) -> None:
        """Create a polygon object with the possibility to fix its color.

        Here, fixing a color means that `set_color`, `set_facecolor`, and `set_edgecolor` will have
        no effect.

        :param axes: The axes object that is used to draw the polygon.
        :param xdata: The x-coordinates of the polygon.
        :param ydata: The y-coordinates of the polygon.
        :fixed_color: Whether or not the color should be fixed (default=False).
        :kwargs: Any additional arguments that will be passed to matplotlib's polygon functionality.
        """
        self.xdata = xdata
        self.ydata = ydata
        self.patch = patches.Polygon(np.array([xdata, ydata]).T, **kwargs)
        axes.add_patch(self.patch)
        self.fixed_color = fixed_color

    def get_xdata(self) -> np.ndarray:
        """Get the x-coordinates of the vertices of the polygon.

        :return: Numpy array with the x-coordinates of the vertices of the polygon.
        """
        return self.xdata

    def get_ydata(self) -> np.ndarray:
        """Get the y-coordinates of the vertices of the polygon.

        :return: Numpy array with the y-coordinates of the vertices of the polygon.
        """
        return self.ydata

    def get_xy(self) -> np.ndarray:
        """Get the (x,y)-coordinates of the vertices of the polygon.

        :return: Numpy array (N-by-2) with the (x,y)-coordinates of the vertices of the polygon.
        """
        return np.array([self.xdata, self.ydata]).T

    def set_xdata(self, xdata: np.ndarray) -> None:
        """Set the x-coordinates of the vertices of the polygon.

        :param xdata: Numpy array with the x-coordinates of the vertices of the polygon.
        """
        self.xdata = xdata
        self.set_xy(np.array([self.xdata, self.ydata]).T)

    def set_ydata(self, ydata: np.ndarray) -> None:
        """Set the y-coordinates of the vertices of the polygon.

        :param ydata: Numpy array with the y-coordinates of the vertices of the polygon.
        """
        self.ydata = ydata
        self.set_xy(np.array([self.xdata, self.ydata]).T)

    def set_xy(self, xydata: np.ndarray) -> None:
        """Set the (x,y)-coordinates of the vertices of the polygon.

        :param xydata: Numpy array (N-by-2) with the (x,y)-coordinates of the vertices of the
                       polygon.
        """
        self.xdata = xydata[:, 0]
        self.ydata = xydata[:, 1]
        self.patch.set_xy(xydata)

    def set_color(self, color: Tuple[float, float, float]) -> None:
        """Set the color of the polygon.

        If the polygon has a fixed color, nothing is done.

        :param color: The color of the polygon.
        """
        if not self.fixed_color:
            self.patch.set_color(color)

    def set_facecolor(self, color: Tuple[float, float, float]) -> None:
        """Set the face color of the polygon.

        If the polygon has a fixed color, nothing is done.

        :param color: The face color of the polygon.
        :param force: Set to True if color must be changed, even if it is fixed.
        """
        if not self.fixed_color:
            self.patch.set_facecolor(color)

    def set_facecolor_forced(self, color: Tuple[float, float, float]) -> None:
        """Set the face color of the polygon, regardless of whether the color is fixed.

        :param color: The face color of the polygon.
        """
        self.patch.set_facecolor(color)

    def set_edgecolor(self, color: Tuple[float, float, float]) -> None:
        """Set the edge color of the polygon.

        If the polygon has a fixed color, nothing is done.

        :param color: The edge color of the polygon.
        """
        if not self.fixed_color:
            self.patch.set_edgecolor(color)

    def remove(self) -> None:
        """Remove the patch."""
        self.patch.remove()



"""Scripts for testing all functionalities from vertex.py.

Author(s): Erwin de Gelder
"""

from traffic_scene_renderer import Vertex, VertexOptions


def test_vertex_creation() -> None:
    vertex = Vertex(0, 0, 0)
    assert str(vertex) == "Vertex[ID=0, x=0.00, y=0.00]"


def test_vertex_creation_with_latlon() -> None:
    vertex_options = VertexOptions(latlon=True)
    # Location of TNO Automotive :)
    vertex = Vertex(0, 51.474457, 5.623897, options=vertex_options)
    assert str(vertex) == "Vertex[ID=0, x=682219.44, y=5705853.69]"


def test_vertex_get_xy() -> None:
    vertex = Vertex(0, 1, 2)
    assert vertex.get_xy() == [1, 2]



"""Truck object.

Author(s): Erwin de Gelder
"""

from typing import Optional, Tuple, Union

import numpy as np
from matplotlib.axes import Axes

from .polygon import Polygon
from .utilities import rotate
from .vehicle import Vehicle, VehicleOptions


class TruckOptions(VehicleOptions):
    """The default values of the options of a truck.

    The following list shows the options (within parentheses default values):
        length (4.5): The length of the truck.
        width (2.5): The width of the truck.
        line_width (1): The line width for plotting the truck.
        x_position_init (0): The initial horizontal position of the truck.
        y_position_init (0): The initial vertical position of the truck.
        angle_init (0): The initial angle of the truck.
        color ((0, 0.4375, 0.75)): Main color of the truck.
        color2 ((0.02, 0.02, 0.02)): Secondary color of the truck.
        color3 ((0.2, 0.2, 0.2)): Tertiary color of the truck.
        trailer (False): Whether the truck has a trailer.
        l_trailer (9.0): Distance front trailer towards pivot point.
        w_trailer (None): Width of trailer (by default, same as width truck).
        l_pivot_truck (1.2): Distance center truck towards pivot point.
        l_pivot_trailer (0.7): Distance front trailer towards pivot point.
        edgecolor ((0, 0, 0)): In case of a non-transparent truck, the color of
            the lines is set using this option.
        layer (2): The layer in which the truck will be plotted.
    """

    length: float = 4.5
    width: float = 2.5
    color2: Tuple[float, float, float] = (0.02, 0.02, 0.02)
    color3: Tuple[float, float, float] = (0.2, 0.2, 0.2)

    trailer: bool = False
    l_trailer: float = 9.0
    w_trailer: float = 0
    l_pivot_truck: float = 1.2
    l_pivot_trailer: float = 0.7

    def __init__(self, **kwargs: Union[bool, float, Tuple[float, float, float]]) -> None:
        """Class containing all kinds of options for a truck.

        :param kwargs: Any options can be set through kwargs.
        """
        VehicleOptions.__init__(self, **kwargs)
        if self.w_trailer == 0:
            self.w_trailer = self.width


class Truck(Vehicle):
    """Plot a truck on the scene.

    Create a truck object using Truck(axes, options).

    Attributes:
        options (TruckOptions): All options. For a detailed description, see above.
        trailer_angle (float): The angle of the trailer wrt the truck.
        axes (Axes): The axes that is used for plotting.
    """

    def __init__(self, axes: Axes, options: Optional[TruckOptions] = None) -> None:
        """Create a truck, possibly with a trailer.

        :param axes: The axes on which the truck must be plotted.
        :param options: Any options for configuring the truck.
        """
        if options is None:
            options = TruckOptions()
        self.trailer_angle = 0.0
        Vehicle.__init__(self, axes, options)
        self.options: TruckOptions

    def plot_vehicle(self) -> None:
        """Plot the truck on the axes."""
        # Plot the truck.
        yoffset = 160.5
        yscale = 301
        xoffset = 103
        xscale = 170

        # Fill with main color, mirror.
        xdata = np.array([103, 32, 32, 19, 17, 18, 21, 22, 25, 30, 38, 47, 63, 103])
        ydata = np.array([152, 152, 156, 156, 153, 86, 49, 38, 24, 18, 13, 11, 10, 10])
        xdata = (np.concatenate((xdata, 2 * xoffset - np.flipud(xdata))) - xoffset) / xscale
        ydata = (np.concatenate((ydata, np.flipud(ydata))) - yoffset) / yscale
        self.fills += (
            Polygon(
                self.axes,
                xdata * self.options.width,
                -ydata * self.options.length,
                facecolor=self.options.color,
                edgecolor=self.options.edgecolor,
                zorder=self.options.layer,
            ),
        )

        # Fill fuel tank, both sides.
        xdata = (np.array([28, 28, 65, 65]) - xoffset) / xscale
        ydatas = [
            np.array([171, 182, 182, 171]),
            np.array([184, 203, 203, 184]),
            np.array([205, 224, 224, 205]),
            np.array([226, 237, 237, 226]),
        ]
        for ydata in ydatas:
            ydata_new = (ydata - yoffset) / yscale
            self.fills += (
                Polygon(
                    self.axes,
                    xdata * self.options.width,
                    -ydata_new * self.options.length,
                    facecolor=self.options.color,
                    edgecolor=self.options.edgecolor,
                    zorder=self.options.layer,
                ),
                Polygon(
                    self.axes,
                    -xdata * self.options.width,
                    -ydata_new * self.options.length,
                    facecolor=self.options.color,
                    edgecolor=self.options.edgecolor,
                    zorder=self.options.layer,
                ),
            )

        # Fill mirror and other black (color2) part, both sides.
        xdatas = [np.array([21, 7, 7, 11, 22]), np.array([37, 37, 58, 58])]
        ydatas = [np.array([49, 45, 41, 38, 38]), np.array([152, 168, 168, 152])]
        for xdata, ydata in zip(xdatas, ydatas):
            xdata_new = (xdata - xoffset) / xscale
            ydata_new = (ydata - yoffset) / yscale
            self.fills += (
                Polygon(
                    self.axes,
                    xdata_new * self.options.width,
                    -ydata_new * self.options.length,
                    facecolor=self.options.color2,
                    edgecolor=self.options.edgecolor,
                    linewidth=self.options.line_width,
                    zorder=self.options.layer,
                    fixed_color=True,
                ),
                Polygon(
                    self.axes,
                    -xdata_new * self.options.width,
                    -ydata_new * self.options.length,
                    facecolor=self.options.color2,
                    edgecolor=self.options.edgecolor,
                    linewidth=self.options.line_width,
                    zorder=self.options.layer,
                    fixed_color=True,
                ),
            )

        # Draw a line on top of the truck.
        xdata = (np.array([40, 42, 164, 166]) - xoffset) / xscale
        ydata = (np.array([152, 106, 106, 152]) - yoffset) / yscale
        self.plots += (
            self.axes.plot(
                xdata * self.options.width,
                -ydata * self.options.length,
                color=self.options.edgecolor,
                linewidth=self.options.line_width,
                zorder=self.options.layer,
            )[0],
        )

        # Draw the contour of the back of the truck, mirror.
        xdata = np.array(
            [
                63,
                63,
                74,
                74,
                27,
                27,
                74,
                74,
                27,
                27,
                74,
                74,
                27,
                27,
                74,
                74,
                30,
                28,
                26,
                26,
                27,
                30,
                30,
                70,
                70,
                73,
                73,
                104,
            ]
        )
        ydata = np.array(
            [
                152,
                165,
                165,
                182,
                182,
                184,
                184,
                203,
                203,
                205,
                205,
                224,
                224,
                226,
                226,
                241,
                241,
                244,
                252,
                300,
                304,
                305,
                311,
                311,
                307,
                307,
                314,
                314,
            ]
        )
        xdata3 = (np.concatenate((xdata, 2 * xoffset - np.flipud(xdata))) - xoffset) / xscale
        ydata3 = (np.concatenate((ydata, np.flipud(ydata))) - yoffset) / yscale
        self.plots += (
            self.axes.plot(
                xdata3 * self.options.width,
                -ydata3 * self.options.length,
                color=self.options.edgecolor,
                linewidth=self.options.line_width,
                zorder=self.options.layer,
            )[0],
        )

        # Draw part to which trailer could be attached (color3), mirror.
        xdata2 = np.array([103, 102, 100, 99, 96, 94, 91, 89, 84, 80, 80, 83, 88, 93, 99, 103])
        ydata2 = np.array(
            [262, 263, 268, 277, 286, 288, 288, 286, 276, 262, 250, 241, 235, 232, 231, 231]
        )
        xdata3 = (np.concatenate((xdata2, 2 * xoffset - np.flipud(xdata2))) - xoffset) / xscale
        ydata3 = (np.concatenate((ydata2, np.flipud(ydata2))) - yoffset) / yscale
        self.fills += (
            Polygon(
                self.axes,
                xdata3 * self.options.width,
                -ydata3 * self.options.length,
                facecolor=self.options.color3,
                edgecolor=self.options.edgecolor,
                linewidth=self.options.line_width,
                zorder=self.options.layer,
                fixed_color=True,
            ),
        )

        # Draw back part of the truck, twice.
        xdata = np.concatenate(
            (xdata, xdata2, np.array([103, 80, 80, 98, 98, 80, 80, 98, 98, 103, 103]))
        )
        ydata = np.concatenate(
            (ydata, ydata2, np.array([228, 228, 205, 205, 194, 194, 170, 170, 228, 228, 152]))
        )
        xdata = (np.concatenate((xdata, 2 * xoffset - np.flipud(xdata))) - xoffset) / xscale
        ydata = (np.concatenate((ydata, np.flipud(ydata))) - yoffset) / yscale
        self.fills += (
            Polygon(
                self.axes,
                xdata * self.options.width,
                -ydata * self.options.length,
                facecolor=self.options.color2,
                linewidth=0,
                zorder=self.options.layer,
                fixed_color=True,
            ),
        )

        # Plot the trailer.
        if self.options.trailer:
            y_truck = (
                self.options.l_pivot_trailer
                - self.options.l_pivot_truck
                - self.options.l_trailer / 2
            )
            xdata = np.array([-1, 1, 1, -1]) * self.options.w_trailer / 2
            ydata = np.array([1, 1, -1, -1]) * self.options.l_trailer / 2 + y_truck
            self.fills += (
                Polygon(
                    self.axes,
                    xdata,
                    ydata,
                    facecolor=self.options.color,
                    edgecolor=self.options.edgecolor,
                    linewidth=self.options.line_width,
                    zorder=self.options.layer + 1,
                ),
            )

    def change_trailer_angle(self, angle: float) -> None:
        """Change the angle of the trailer.

        :param angle: The angle of the trailer wrt the truck.
        """
        x_pivot = self.position.x_center - np.sin(self.position.angle) * self.options.l_pivot_truck
        y_pivot = self.position.y_center - np.cos(self.position.angle) * self.options.l_pivot_truck
        xy_data = self.fills[-1].get_xy() - [x_pivot, y_pivot]
        xy_data[:, 0], xy_data[:, 1] = rotate(
            xy_data[:, 0], xy_data[:, 1], self.position.angle + self.trailer_angle
        )
        xy_data[:, 0], xy_data[:, 1] = rotate(
            xy_data[:, 0], xy_data[:, 1], -self.position.angle - angle
        )
        xy_data[:, 0] += x_pivot
        xy_data[:, 1] += y_pivot
        self.fills[-1].set_xy(xy_data)
        self.trailer_angle = angle



"""Scripts for testing traffic_light.py.

Author(s): Erwin de Gelder
"""

from pathlib import Path

import matplotlib as mpl
import matplotlib.pyplot as plt
import pytest

from traffic_scene_renderer import (
    NoAmberError,
    TrafficLight,
    TrafficLightOptions,
    TrafficLightStatus,
)

from .test_static_objects import save_fig

mpl.use("Agg")


def test_traffic_light_creation() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-1, 1)
    axes.set_ylim(-2, 2)
    TrafficLight(axes, TrafficLightOptions(radius=0.5))
    save_fig(fig, axes, Path("traffic_light") / "idle.png", 4)


def test_traffic_light_without_amber() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-50, 50)  # Set large, in order to get a large traffic light.
    axes.set_ylim(-2, 2)
    TrafficLight(axes, TrafficLightOptions(amber=False))
    axes.set_xlim(-1, 1)
    save_fig(fig, axes, Path("traffic_light") / "no_amber.png", 4)


def test_traffic_light_status() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-4, 4)
    axes.set_ylim(-2, 2)
    traffic_light = TrafficLight(axes, TrafficLightOptions(radius=0.5))
    traffic_light.change_pos(-3, 0)
    traffic_light.set_status(TrafficLightStatus.IDLE)
    traffic_light = TrafficLight(axes, TrafficLightOptions(radius=0.5))
    traffic_light.change_pos(-1, 0)
    traffic_light.set_status(TrafficLightStatus.RED)
    traffic_light = TrafficLight(axes, TrafficLightOptions(radius=0.5))
    traffic_light.change_pos(1, 0)
    traffic_light.set_status(TrafficLightStatus.AMBER)
    traffic_light = TrafficLight(axes, TrafficLightOptions(radius=0.5))
    traffic_light.change_pos(3, 0)
    traffic_light.set_status(TrafficLightStatus.GREEN)
    traffic_light = TrafficLight(axes, TrafficLightOptions(radius=0.5))
    traffic_light.set_status(TrafficLightStatus.RED)
    traffic_light.set_status(TrafficLightStatus.REMOVED)
    save_fig(fig, axes, Path("traffic_light") / "different_status.png", 8)


def test_traffic_light_status_no_amber() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-2, 2)
    axes.set_ylim(-2, 2)
    traffic_light = TrafficLight(axes, TrafficLightOptions(radius=0.5, amber=False))
    traffic_light.change_pos(-1, 0)
    traffic_light.set_status(TrafficLightStatus.RED)
    traffic_light = TrafficLight(axes, TrafficLightOptions(radius=0.5, amber=False))
    traffic_light.change_pos(1, 0)
    traffic_light.set_status(TrafficLightStatus.GREEN)
    save_fig(fig, axes, Path("traffic_light") / "no_amber_status.png", 8)


def test_no_amber_error() -> None:
    fig, axes = plt.subplots()
    traffic_light = TrafficLight(axes, TrafficLightOptions(amber=False))
    try:
        traffic_light.set_status(TrafficLightStatus.AMBER)
    except NoAmberError:
        pass
    else:
        pytest.fail("NoAmberError should be raised.")
    plt.close(fig)


def test_str() -> None:
    fig, axes = plt.subplots()
    traffic_light = TrafficLight(axes)
    assert str(traffic_light) == "Traffic light, status=IDLE"
    traffic_light.set_status(TrafficLightStatus.REMOVED)
    assert str(traffic_light) == "Traffic light, status=REMOVED"
    traffic_light.set_status(TrafficLightStatus.RED)
    assert str(traffic_light) == "Traffic light, status=RED"
    traffic_light.set_status(TrafficLightStatus.AMBER)
    assert str(traffic_light) == "Traffic light, status=AMBER"
    traffic_light.set_status(TrafficLightStatus.GREEN)
    assert str(traffic_light) == "Traffic light, status=GREEN"



"""Scripts for testing way.py.

Author(s): Erwin de Gelder
"""

from pathlib import Path
from typing import List, Union

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pytest
from matplotlib.axes import Axes

from traffic_scene_renderer import Crossing, IndexVertexError, Vertex, Way, WayOptions

from .test_static_objects import save_fig

mpl.use("Agg")


def plot_way(axes: Axes, way: Way) -> None:
    axes.plot(way.parms.left.x_data, way.parms.left.y_data, color=way.options.side_color)
    axes.plot(way.parms.right.x_data, way.parms.right.y_data, color=way.options.side_color)
    if way.parms.plot.lines:
        lines = np.array(way.parms.plot.lines)
        axes.plot(
            [lines[:, 0], lines[:, 2]], [lines[:, 1], lines[:, 3]], color=way.options.line_color
        )
    way.plot_markers(axes)


def test_way_creation() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-5, 5)
    way = Way([Vertex(0, -10, 0), Vertex(1, 10, 0)], WayOptions(line_color=(1, 0, 0)))
    way.process()
    plot_way(axes, way)
    save_fig(fig, axes, Path("way") / "straight.png", 10)
    assert str(way) == "Way[vertices=[0, 1]]"


def test_way_different_types() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-20, 20)
    axes.set_ylim(-30, 20)
    ways = (
        Way([Vertex(0, -20, 18), Vertex(1, 20, 18)], WayOptions(highway="footway")),
        Way(
            [Vertex(2, -20, 13), Vertex(3, 20, 13)],
            WayOptions(maxspeed=40, line_color=(0.5, 0, 0), side_color=(0.5, 0, 0)),
        ),
        Way(
            [Vertex(4, -20, 5), Vertex(5, 20, 5)],
            WayOptions(maxspeed=60, line_color=(0, 0.5, 0), side_color=(0, 0.5, 0)),
        ),
        Way(
            [Vertex(6, -20, -4), Vertex(7, 20, -4)],
            WayOptions(maxspeed=90, line_color=(0, 0, 0.5), side_color=(0, 0, 0.5)),
        ),
        Way(
            [Vertex(8, -20, -14), Vertex(9, 20, -14)],
            WayOptions(maxspeed=110, line_color=(0, 0.5, 0.5), side_color=(0, 0.5, 0.5)),
        ),
        Way(
            [Vertex(10, -20, -24), Vertex(11, 20, -24)],
            WayOptions(
                maxspeed=110, line_color=(0.5, 0.5, 0), side_color=(0.5, 0.5, 0), linetype="solid"
            ),
        ),
    )
    for way in ways:
        way.process()
        plot_way(axes, way)
    axes.text(-8, 18, "Footway")
    axes.text(-8, 13, "Urban road")
    axes.text(-8, 5, "Interurban road")
    axes.text(-8, -4, 'Motorway "low speed"')
    axes.text(-8, -14, 'Motorway "high speed"')
    axes.text(-8, -24, "Motorway solid line")
    save_fig(fig, axes, Path("way") / "different_ways.png", 20)


def test_split_way() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-5, 5)
    way_orig = Way(
        [Vertex(0, -10, 0), Vertex(1, 0, 0), Vertex(2, 10, 0)], WayOptions(line_color=(0, 0, 0))
    )
    way_split = way_orig.split(1)
    way_split.options.line_color = (1, 0, 0)
    way_split.options.side_color = (1, 0, 0)
    for way in (way_orig, way_split):
        way.process()
        plot_way(axes, way)
    save_fig(fig, axes, Path("way") / "split_way.png", 10)


def test_insert_and_pop_vertex() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-10, 10)
    vertex_left = Vertex(0, -10, 6)
    vertex_right = Vertex(1, 10, 6)
    way = Way([vertex_left, vertex_right], WayOptions(line_color=(0, 0, 0)))
    assert way.insert_vertex(Vertex(2, 0, 5), 1) == (False, False)
    way.process()
    plot_way(axes, way)
    assert way.pop_vertex(1) == (False, False)
    vertex_left.ycoordinate = -5
    vertex_right.ycoordinate = -5
    way.process()  # Reprocessing is necessary.
    plot_way(axes, way)
    save_fig(fig, axes, Path("way") / "insert_and_pop_vertex.png", 10)


def test_insert_and_pop_vertex_in_between_crossing() -> None:
    vertices = [
        Vertex(0, -10, 0),
        Vertex(1, 0, 0),
        Vertex(2, 0, 10),
        Vertex(3, 10, 0),
        Vertex(4, 10, 10),
        Vertex(5, 20, 0),
    ]
    ways = [
        Way([vertices[0], vertices[1]]),
        Way([vertices[1], vertices[2]]),
        Way([vertices[3], vertices[1]]),
        Way([vertices[3], vertices[4]]),
        Way([vertices[3], vertices[5]]),
    ]
    Crossing(0, vertices[1], ways[:3])
    Crossing(1, vertices[3], ways[2:])
    assert ways[2].insert_vertex(Vertex(6, 5, 1), 1) == (True, True)
    assert ways[2].pop_vertex(1) == (True, True)


def test_warning_lambda() -> None:
    vertices = [
        Vertex(0, -10, 0),
        Vertex(1, 0, 0),
        Vertex(2, 0, 10),
        Vertex(3, 1, 0),
        Vertex(4, 10, 0),
        Vertex(5, 20, 10),
        Vertex(6, 20, 0),
    ]
    ways = [
        Way([vertices[0], vertices[1]]),
        Way([vertices[1], vertices[2]]),
        Way([vertices[1], vertices[3], vertices[4]]),
        Way([vertices[4], vertices[5]]),
        Way([vertices[4], vertices[6]]),
    ]
    crossing1 = Crossing(0, vertices[1], ways[:3])
    crossing2 = Crossing(1, vertices[4], ways[2:])
    crossing1.process()
    crossing2.process()
    with pytest.warns(UserWarning):
        ways[2].get_xy()
    vertices[3].xcoordinate = 9
    crossing1.process()
    crossing2.process()
    with pytest.warns(UserWarning):
        ways[2].get_xy()


def test_vertex_invalid_index_error() -> None:
    way = Way([Vertex(0, 0, 0), Vertex(1, 10, 0)])
    try:
        way.insert_vertex(Vertex(2, 5, 0), -2)
    except IndexVertexError:
        pass
    else:
        pytest.fail("IndexInsertVertexError should be raised.")
    try:
        way.pop_vertex(0)
    except IndexVertexError:
        pass
    else:
        pytest.fail("IndexInsertVertexError should be raised.")


def test_set_nlanes() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-5, 5)
    way = Way([Vertex(0, -10, 0), Vertex(1, 10, 0)], WayOptions(line_color=(1, 0, 0), lanewidth=2))
    way.set_nlanes(4)
    way.process()
    plot_way(axes, way)
    save_fig(fig, axes, Path("way") / "fourlanes.png", 10)


def plot_way_with_offset(axes: Axes, y_orig: int, offset: Union[float, List[float]]) -> None:
    way = Way(
        [Vertex(0, -10, y_orig), Vertex(1, 0, y_orig), Vertex(2, 10, y_orig)],
        WayOptions(line_color=(0.9, 0.9, 0.9), side_color=(0.9, 0.9, 0.9)),
    )
    way.process()
    plot_way(axes, way)
    way.apply_offset(offset)
    way.process()
    way.options.side_color = (0, 0, 0)
    way.options.line_color = (0, 0, 0)
    plot_way(axes, way)


def test_apply_offset() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-12, 12)
    plot_way_with_offset(axes, 7, [0.5, 1, 2])
    plot_way_with_offset(axes, -1, [0.5, 2])
    plot_way_with_offset(axes, -8, 1)
    save_fig(fig, axes, Path("way") / "apply_offset.png", 10)


def test_road_markers() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-5, 5)
    way = Way(
        [Vertex(0, -10, 0), Vertex(1, 0, 0), Vertex(2, 10, 0)],
        WayOptions(line_color=(1, 0, 0), turnlanes="none|right", marker_border_color=(0, 0, 0.5)),
    )
    way.process()
    plot_way(axes, way)
    way.plot_markers(axes)
    save_fig(fig, axes, Path("way") / "way_with_markers.png", 10)


def test_warning_markers_vs_nr_of_lanes() -> None:
    fig, axes = plt.subplots()
    way = Way([Vertex(0, -10, 0), Vertex(1, 10, 0)], WayOptions(turnlanes="through", nlanes=2))
    with pytest.warns(UserWarning):
        way.plot_markers(axes)
    plt.close(fig)


def test_get_n_markers_without_xy_data() -> None:
    way = Way([Vertex(0, -10, 0), Vertex(1, 10, 0)], WayOptions(marker_interval=100))
    assert way.get_n_markers() == 1


def test_plot_marker_single_lane() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-5, 5)
    way = Way(
        [Vertex(0, -10, 0), Vertex(1, 10, 0)],
        WayOptions(line_color=(1, 0, 0), n_markers=8, marker_border_color=(0, 0, 0.5)),
    )
    way.plot_markers_lane(axes, 1, "through")
    way.plot_markers_lane(axes, 0, "rev_through")
    way.process()
    plot_way(axes, way)
    save_fig(fig, axes, Path("way") / "way_with_reversed_markers.png", 10)


def test_compute_position_markers() -> None:
    way = Way(
        [Vertex(0, -10, 0), Vertex(1, 10, 0)],
        WayOptions(line_color=(1, 0, 0), n_markers=2, nlanes=1),
    )
    easting, northing, angle = way.compute_position_markers(0)
    assert easting[0] == pytest.approx(-5.0)
    assert easting[1] == pytest.approx(5.0)
    assert northing[0] == pytest.approx(0.0)
    assert northing[1] == pytest.approx(0.0)
    assert angle[0] == pytest.approx(np.pi / 2)
    assert angle[1] == pytest.approx(np.pi / 2)


def test_non_unique_vertex_ids() -> None:
    with pytest.warns(UserWarning):
        Way([Vertex(0, 0, 0), Vertex(0, 10, 0)])



"""Constructing ways.

Author(s): Erwin de Gelder
"""

import copy
import warnings
from typing import List, Optional, Tuple, Union

import numpy as np
from matplotlib.axes import Axes

from .options import Options
from .static_objects import TurnArrow, TurnArrowOptions
from .vertex import Vertex

DEFAULT_LANE_WIDTH = 2.8  # [m]
FOOTWAY_LANE_WIDTH = 1.5  # [m]
INTERURBAN_LANE_WIDTH = 3  # [m]
HIGHWAY_LANE_WIDTH = 3.5  # [m] Default for Dutch motorways
MAX_URBAN_SLOW_SPEED = 30
MAX_URBAN_SPEED = 50
MAX_INTERURBAN_SPEED = 80
MAX_MOTORWAY_SLOW_SPEED = 100


class WayOptions(Options):
    """The default values of the options of a way object."""

    highway: Optional[str] = None
    maxspeed: Optional[float] = None
    nlanes: int = 0
    lanewidth: float = -1
    layer: int = 0
    oneway: bool = False
    turnlanes: Optional[str] = None
    linetype: Optional[str] = None
    line_color: Tuple[float, float, float] = (1, 1, 1)
    marker_border_color: Tuple[float, float, float] = (1, 1, 1)
    show_border: bool = True
    marker_interval: float = 10
    n_markers: Optional[int] = None
    marker_fill_color: Tuple[float, float, float] = (1, 1, 1)
    color: Tuple[float, float, float] = (-1, -1, -1)
    side_color: Tuple[float, float, float] = (-1, -1, -1)
    line_interval: float = -1
    line_length: float = -1

    def __init__(self, **kwargs: Union[bool, float, str, Tuple[float, float, float]]) -> None:
        """Create a container for all options for a way.

        :param kwargs: Use kwargs to set any option other then the default option.
        """
        Options.__init__(self, **kwargs)
        if self.lanewidth == -1:
            self.lanewidth = self._compute_lane_width()
        if self.nlanes == 0:
            self.nlanes = self._compute_nlanes()
        if self.line_interval == -1 or self.line_length == -1:
            # They both need to be set, otherwise they are calculated.
            self.line_interval, self.line_length = self._compute_line_info()
        if self.color == (-1, -1, -1):
            self.color = self._determine_color()
        if self.side_color == (-1, -1, -1):
            self.side_color = self._determine_side_color()

    def _compute_lane_width(self) -> float:
        """Determine the lanewidth, based on the maxspeed.

        :return: lane width.
        """
        if self.maxspeed is None:
            if self.highway == "footway":
                return FOOTWAY_LANE_WIDTH
            return DEFAULT_LANE_WIDTH
        if self.maxspeed <= MAX_URBAN_SPEED:
            return DEFAULT_LANE_WIDTH
        if self.maxspeed <= MAX_INTERURBAN_SPEED:
            return INTERURBAN_LANE_WIDTH
        # For simplicity, assume highway
        return HIGHWAY_LANE_WIDTH

    def _compute_nlanes(self) -> int:
        """Determine the number of lanes; always 2, unless it is a footway (then 1).

        :return: Number of lanes.
        """
        if self.highway == "footway":
            return 1
        return 2  # Assume it is a bidirectional road

    def _compute_line_info(self) -> Tuple[float, float]:
        """Compute line information.

        :return: line interval and line length.
        """
        if self.maxspeed is None or self.maxspeed <= MAX_URBAN_SLOW_SPEED:
            line_interval, line_length = 3.0, 1.0
        elif self.maxspeed <= MAX_URBAN_SPEED:
            line_interval, line_length = 5.0, 5 / 3
        elif self.maxspeed <= MAX_INTERURBAN_SPEED:
            line_interval, line_length = 8.0, 8 / 3
        elif self.maxspeed <= MAX_MOTORWAY_SLOW_SPEED:
            line_interval, line_length = 10.0, 10 / 3
        else:
            line_interval, line_length = 12.0, 4.0

        if self.linetype == "solid":
            # It is a dirty solution, but we make the line interval similar to the line length.
            return line_interval, line_interval
        return line_interval, line_length

    def _determine_color(self) -> Tuple[float, float, float]:
        """Determine the color of the road.

        :return: RGB tuple.
        """
        if self.highway == "footway":
            return 1, 0.9, 0.8
        return 0.8, 0.8, 0.8

    def _determine_side_color(self) -> Tuple[float, float, float]:
        """Determine the side color of the road.

        :return: RGB tuple.
        """
        if self.highway == "footway":
            return 0.8, 1, 0.8
        return 0, 0, 0


class WayConnection(Options):
    """Class containing parameters of a way object concerning its connections."""

    i_start: int = -1
    i_end: int = -1
    x_left_start: float = 0
    x_right_start: float = 0
    y_left_start: float = 0
    y_right_start: float = 0
    x_left_end: float = 0
    x_right_end: float = 0
    y_left_end: float = 0
    y_right_end: float = 0


class WayCrossing(Options):
    """Class containing parameters of a way object concerning its crossings."""

    i_start: int = -1
    i_end: int = -1
    start_lambda_left: float = 0
    start_lambda_right: float = 0
    end_lambda_left: float = 0
    end_lambda_right: float = 0


class XYData(Options):
    """Class containing xy-data that is used for way objects."""

    def __init__(self, **kwargs: np.ndarray) -> None:
        """Create an object with x_data and y_data stored in numpy array.

        :param kwargs: x_data and y_data may be already provided.
        """
        self.x_data = np.array([])
        self.y_data = np.array([])
        Options.__init__(self, **kwargs)


class WayPlotData(Options):
    """Class containing plot data of a way object."""

    i_interval: int = 0

    def __init__(self, **kwargs: Union[List, List[TurnArrow], np.ndarray]) -> None:
        """Create an object to store all plotting information of a way.

        :param kwargs: Any parameters that are to be set can be passed via kwargs.
        """
        self.lines: List = []
        self.arrows: List[TurnArrow] = []  # For turning arrows.
        self.lengths = np.array([])
        self.xyline = np.array([])
        Options.__init__(self, **kwargs)


class WayParameters(Options):
    """Class containing parameters of a way object."""

    hwidth: float = 0

    def __init__(
        self, **kwargs: Union[float, List[float], WayConnection, WayCrossing, XYData, WayPlotData]
    ) -> None:
        """Create an object to store all parameters for a way.

        :param kwargs: Any parameters that are to be set can be passed via kwargs.
        """
        self.offset: List[float] = []
        self.connection: WayConnection = WayConnection()
        self.crossing: WayCrossing = WayCrossing()
        self.left: XYData = XYData()
        self.right: XYData = XYData()
        self.plot: WayPlotData = WayPlotData()
        Options.__init__(self, **kwargs)


class IndexVertexError(Exception):
    """Error to be raised in case the index of the vertex is invalid."""

    def __init__(self, index: int, maxindex: int) -> None:
        """Description of error."""
        super().__init__(f"Index should be in range 1 to {maxindex}, but is {index}'.")


class Way:
    """A way of a road network.

    Attributes:
        options (WayOptions): All options. For a detailed description, see above.
        vertices (List[Vertex]): All vertices of the way.
        ivs (List[int]): The indices of all vertices of the way.
        parms (WayParameters): All parameters that are used for rendering the way.
    """

    def __init__(self, vertices: List[Vertex], options: Optional[WayOptions] = None) -> None:
        """Initializes a Way object.

        :param vertices: List of vertices of the way.
        :param options: possible options
        """
        self.options = WayOptions() if options is None else options
        self.vertices = vertices

        # Make list of indices of the vertices.
        self.ivs = [v.idx for v in self.vertices]

        # Check if vertices have a unique ID.
        if len(self.ivs) != len(set(self.ivs)):
            warnings.warn(
                "The ID of the vertices of this way are not unique. "
                "This may lead to strange results.",
                stacklevel=2,
            )

        self.parms = WayParameters()
        self.parms.offset = [0.0 for _ in self.ivs]  # Offset per node, right is positive.
        self._compute_hwidth()

    def split(self, index: int) -> "Way":
        """Split way into two ways.

        :param index: index of vertex that belong to new way.
        :return: way that is cut off.
        """
        way = copy.deepcopy(self)
        self.vertices = self.vertices[: index + 1]
        way.vertices = way.vertices[index:]
        self.ivs = self.ivs[: index + 1]
        way.ivs = way.ivs[index:]
        self.parms.offset = self.parms.offset[: index + 1]
        way.parms.offset = way.parms.offset[index:]
        return way

    def insert_vertex(self, vertex: Vertex, index: int) -> Tuple[bool, bool]:
        """Insert a vertex.

        :param vertex: Vertex to be inserted.
        :param index: Index of vertex of the way where the given index will be inserted before.
        :return: Two booleans, tellings whether the start or end crossings need to be processed
                 again.
        """
        # Convert index to a positive number on the interval [0, len(self.ivs)-1).
        # This is done, such that the index refers to the index that the new vertex will have
        # (this is not the case if index=-1).
        if index < 0:
            index += len(self.ivs)
        if index == 0 or abs(index) >= len(self.ivs):
            raise IndexVertexError(index, len(self.ivs) - 1)
        self.vertices.insert(index, vertex)
        self.ivs.insert(index, vertex.idx)
        dist1 = np.hypot(
            self.vertices[index - 1].xcoordinate - self.vertices[index].xcoordinate,
            self.vertices[index - 1].ycoordinate - self.vertices[index].ycoordinate,
        )
        dist2 = np.hypot(
            self.vertices[index].xcoordinate - self.vertices[index + 1].xcoordinate,
            self.vertices[index].ycoordinate - self.vertices[index + 1].ycoordinate,
        )
        self.parms.offset.insert(
            index,
            (self.parms.offset[index - 1] * dist2 + self.parms.offset[index] * dist1)
            / (dist1 + dist2),
        )
        process_start_again = False
        process_end_again = False
        if index == 1 and self.parms.crossing.i_start >= 0:
            process_start_again = True
        if index == len(self.ivs) - 2 and self.parms.crossing.i_end >= 0:
            process_end_again = True
        return process_start_again, process_end_again

    def pop_vertex(self, index: int) -> Tuple[bool, bool]:
        """Remove a vertex from this road.

        Note that this vertex cannot be at the start or at the end of the way.

        :param index: Index of vertex to be removed.
        :return: Two booleans, tellings whether the start or end crossings need to be processed
                 again.
        """
        # Check if index is correct.
        if index == 0 or abs(index) >= len(self.ivs) - 1:
            raise IndexVertexError(index, len(self.ivs) - 1)

        # Check if crossing needs to be processed again. This is the case when the vertex
        # is used for constructing the crossing (i.e. the second or second-last vertex).
        process_start_again = False
        process_end_again = False
        if index % len(self.ivs) == 1 and self.parms.crossing.i_start >= 0:
            process_start_again = True
        if index % len(self.ivs) == len(self.ivs) - 2 and self.parms.crossing.i_end >= 0:
            process_end_again = True

        # Remove vertex
        self.vertices.pop(index)
        self.ivs.pop(index)
        self.parms.offset.pop(index)
        return process_start_again, process_end_again

    def get_xy(self) -> np.ndarray:
        """Get the x-y coordinates of the vertices.

        :return: np.array with x,y coordinates
        """
        # Get (x,y) coordinates. First and last coordinates may differ, depending on whether
        # they are attached to a crossing/connection.
        xy_data = np.array([[v.xcoordinate, v.ycoordinate] for v in self.vertices])
        xy_tmp = xy_data[-2].copy()  # We need this later, but xy[-2] might change, so we store it.
        if self.parms.crossing.i_start >= 0:
            lamb = max(
                self.parms.crossing.start_lambda_left, self.parms.crossing.start_lambda_right
            )
            if lamb > 1:
                warnings.warn(
                    f"Lambda > 1, this will result in faulty roads! (lambda = {lamb:.2f})",
                    stacklevel=2,
                )
            xy_data[0] = xy_data[0] + lamb * (xy_data[1] - xy_data[0])
        if self.parms.crossing.i_end >= 0:
            lamb = max(self.parms.crossing.end_lambda_left, self.parms.crossing.end_lambda_right)
            if lamb > 1:
                warnings.warn(
                    f"Lambda > 1, this will result in faulty roads! (lambda = {lamb:.2f})",
                    stacklevel=2,
                )
            xy_data[-1] = xy_data[-1] + lamb * (xy_tmp - xy_data[-1])

        # Apply offset
        return apply_offset(xy_data, self.parms.offset)

    def set_nlanes(self, nlanes: int) -> None:
        """Set the number of lanes of the way.

        This function needs to be used instead of setting directly self.options.nlanes,
        because this function will recalculate the hwidth.

        :param nlanes: The number of lanes.
        """
        self.options.nlanes = nlanes
        self._compute_hwidth()

    def process(self) -> None:
        """Process way to make it suitable for plotting.

        Note that corresponding crossings need to be processed before processing the road.
        """
        xy_data = self.get_xy()
        xy_left = apply_offset(xy_data, self.parms.hwidth)
        xy_right = apply_offset(xy_data, -self.parms.hwidth)
        self.parms.left.x_data, self.parms.left.y_data = xy_left[:, 0], xy_left[:, 1]
        self.parms.right.x_data, self.parms.right.y_data = xy_right[:, 0], xy_right[:, 1]

        # Check if we need to correct final point, because of a connection.
        if self.parms.connection.i_start >= 0 and self.parms.crossing.i_start == -1:
            self.parms.left.x_data[0] = self.parms.connection.x_left_start
            self.parms.right.x_data[0] = self.parms.connection.x_right_start
            self.parms.left.y_data[0] = self.parms.connection.y_left_start
            self.parms.right.y_data[0] = self.parms.connection.y_right_start
        if self.parms.connection.i_end >= 0 and self.parms.crossing.i_end == -1:
            self.parms.left.x_data[-1] = self.parms.connection.x_left_end
            self.parms.right.x_data[-1] = self.parms.connection.x_right_end
            self.parms.left.y_data[-1] = self.parms.connection.y_left_end
            self.parms.right.y_data[-1] = self.parms.connection.y_right_end

        # Generate lines
        if (
            self.options.nlanes is not None
            and self.options.nlanes > 1
            and self.options.linetype != "none"
        ):
            self.parms.plot.lines = []  # Empty list, even if process() is called again.
            for ilane in range(1, self.options.nlanes):
                self.generate_lane_lines(ilane, xy_data)

    def generate_lane_lines(self, ilane: int, xy_data: np.ndarray) -> None:
        """Compute the coordinates of the line markers of a specific lane.

        The start and the end of road will have a line of half the specified length, e.g.
        # |-   --   --   --   --   -|  (e.g., line_length=2, line_interval=5, nlines=6).
        The line information is stored in self.parms.plot.lines. Hence, nothing is returned.

        :param ilane: Index of the lane, ranging from 1 to the number of lanes minus 1.
        :param xy_data: (x,y) data of the lane, obtained using self.get_xy().
        """
        self.parms.plot.xyline = apply_offset(
            xy_data, self.options.lanewidth * (ilane - self.options.nlanes / 2)
        )
        self.parms.plot.lengths = np.hypot(
            np.diff(self.parms.plot.xyline[:, 0]), np.diff(self.parms.plot.xyline[:, 1])
        )
        total_length = np.sum(self.parms.plot.lengths)
        nlines = np.round(total_length / self.options.line_interval).astype(int) + 1
        interval_ratio = total_length / ((nlines - 1) * self.options.line_interval)

        # Store line in array: [e0, n0, e1, n1] (e=easting, n=northing)
        line = [self.parms.plot.xyline[0, 0], self.parms.plot.xyline[0, 1], 0, 0]
        distance = 0.0  # Distance from starting point
        self.parms.plot.i_interval = 0
        for i in range(nlines):
            # Compute end point of line
            if i in (0, nlines - 1):
                distance += self.options.line_length * interval_ratio / 2
            else:
                distance += self.options.line_length * interval_ratio

            line[2], line[3], distance = self.point_on_line(distance)
            self.parms.plot.lines.append(line.copy())

            # Compute starting point of next line
            if i == nlines - 1:  # No need to do this when there is no next line
                break
            distance += (self.options.line_interval - self.options.line_length) * interval_ratio
            line[0], line[1], distance = self.point_on_line(distance)

    def point_on_line(self, distance: float) -> Tuple[float, float, float]:
        """Compute the point on the line that is a specified distance away from the start.

        The line is defined by the coordinates of self.parms.plot.xy_line. It is
        assumed that this function is only called with increasing distance. If
        this is not the case, some weird results can be produced. If the distance
        is smaller, the index of the interval (self.parms.plot.i_interval) needs
        to be reset to 0.

        :param distance: The distance to the next point.
        :return: The easting and northing position of the new point on the line and the updated
                 distance.
        """
        while distance > self.parms.plot.lengths[self.parms.plot.i_interval]:
            if self.parms.plot.i_interval + 1 == len(self.parms.plot.lengths):
                break
            distance -= self.parms.plot.lengths[self.parms.plot.i_interval]
            self.parms.plot.i_interval += 1
        d_scaled = distance / self.parms.plot.lengths[self.parms.plot.i_interval]
        easting = (
            self.parms.plot.xyline[self.parms.plot.i_interval, 0] * (1 - d_scaled)
            + self.parms.plot.xyline[self.parms.plot.i_interval + 1, 0] * d_scaled
        )
        northing = (
            self.parms.plot.xyline[self.parms.plot.i_interval, 1] * (1 - d_scaled)
            + self.parms.plot.xyline[self.parms.plot.i_interval + 1, 1] * d_scaled
        )
        return easting, northing, distance

    def apply_offset(self, offset: Union[float, List[float]]) -> None:
        """Apply offset to the (x,y) data of the vertices of this way.

        :param offset: Single offset or list of offsets. In case of list of offsets, there are
                       two options.
                       - A list of 2 floats defining the offsets at both ends. Using linear
                         interpolation, the offsets at the other nodes are determined.
                       - A list of N floats. In this case, node i will have an offset of
                         'offset[i]'.
        """
        if isinstance(offset, List):
            if len(offset) == 2:  # noqa: PLR2004
                xy_data = [[vertex.xcoordinate, vertex.ycoordinate] for vertex in self.vertices]
                distance = np.array(
                    [
                        np.hypot(
                            xy_data[i][0] - xy_data[i - 1][0], xy_data[i][1] - xy_data[i - 1][1]
                        )
                        for i in range(1, len(xy_data))
                    ]
                )
                cumlative_dist = np.concatenate(([0], np.cumsum(distance)))
                new_offset = offset[0] + cumlative_dist / cumlative_dist[-1] * (
                    offset[1] - offset[0]
                )
            else:
                new_offset = offset
        else:
            new_offset = [1 for _ in range(len(self.parms.offset))]
        for i in range(len(self.parms.offset)):
            self.parms.offset[i] += new_offset[i]

    def plot_markers(self, axes: Axes) -> None:
        """Plot the markers.

        :param axes: The axes that are used for plotting.
        """
        # If we have no futher information about the lanes, we can't plot any markers.
        if self.options.turnlanes is None:
            return

        # Get directions. If the number of directions is different than the number of lanes,
        # then return.
        directions = self.options.turnlanes.split("|")
        if len(directions) != self.options.nlanes:
            warnings.warn(
                "Does not know how to plot markers: number of directions is different "
                "from number of lanes",
                stacklevel=2,
            )
            return

        xy_data = self.get_xy()
        self.get_n_markers(xy_data)  # This will set self.option.n_markers if not already set.
        for ilane, direction in enumerate(directions):
            self.plot_markers_lane(axes, ilane, direction, xy_data)

    def get_n_markers(self, xy_data: Optional[np.ndarray] = None) -> int:
        """Return the number of markers that are to be plotted.

        This function will set the option n_markers. If this option is already set,
        the function simply returns this value.

        :param xy_data: (x,y)-coordinates of the road. If not provided, it will be computed.
        :return: Number of markers.
        """
        if self.options.n_markers is not None:
            return self.options.n_markers

        if xy_data is None:
            xy_data = self.get_xy()
        lengths = np.hypot(np.diff(xy_data[:, 0]), np.diff(xy_data[:, 1]))
        total_length = np.sum(lengths)
        n_markers = np.floor(total_length / self.options.marker_interval).astype(int)
        if n_markers == 0:
            n_markers = 1
        self.options.n_markers = n_markers
        return n_markers

    def plot_markers_lane(
        self, axes: Axes, ilane: int, direction: str, xy_data: Optional[np.ndarray] = None
    ) -> None:
        """Plot the markers for a single lane.

        :param axes: The axes that are used for plotting.
        :param ilane: The index of the lane.
        :param direction: The direction of the lane marker, e.g., 'left', 'right', 'leftright'.
        :param xy_data: (x,y)-coordinates of the road. If not provided, it will be computed.
        :return:
        """
        if direction == "none" or self.options.n_markers == 0 or not direction:
            return

        if xy_data is None:
            xy_data = self.get_xy()

        easting, northing, angle = self.compute_position_markers(ilane, xy_data)
        options = TurnArrowOptions(
            edge_color=self.options.marker_border_color, face_color=self.options.marker_fill_color
        )
        for i in range(self.get_n_markers()):
            # Create static object
            if direction.startswith("rev_"):
                angle[i] += np.pi
                turn_arrow = TurnArrow(axes, direction=direction[4:], options=options)
            else:
                turn_arrow = TurnArrow(axes, direction=direction, options=options)
            turn_arrow.change_pos(easting[i], northing[i], angle[i])
            self.parms.plot.arrows.append(turn_arrow)

    def compute_position_markers(
        self, ilane: int, xy_data: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Compute the positions of the markers of a single lane.

        :param ilane: The index of the lane.
        :param xy_data: (x,y)-coordinates of the road. If not provided, it will be computed.
        :return: Numpy arrays of the easting, northing, and angle of the markers.
        """
        if xy_data is None:
            xy_data = self.get_xy()

        xyline = apply_offset(
            xy_data, self.options.lanewidth * ((self.options.nlanes - 1) / 2 - ilane)
        )
        lengths = np.hypot(np.diff(xyline[:, 0]), np.diff(xyline[:, 1]))
        total_length = np.sum(lengths)
        marker_interval = total_length / self.options.n_markers
        distance = 0  # Distance from starting point
        i_interval = 0
        easting = np.zeros(self.get_n_markers())
        northing = np.zeros(self.get_n_markers())
        angle = np.zeros(self.get_n_markers())
        for i in range(self.get_n_markers()):
            if i == 0:
                # First and last line have half of the normal length.
                distance += marker_interval / 2
            else:
                distance += marker_interval
            while i_interval < len(lengths) and distance > lengths[i_interval]:
                distance -= lengths[i_interval]
                i_interval += 1
            d_scaled = distance / lengths[i_interval]
            easting[i] = (
                xyline[i_interval, 0] * (1 - d_scaled) + xyline[i_interval + 1, 0] * d_scaled
            )
            northing[i] = (
                xyline[i_interval, 1] * (1 - d_scaled) + xyline[i_interval + 1, 1] * d_scaled
            )
            angle[i] = np.arctan2(
                xyline[i_interval + 1, 0] - xyline[i_interval, 0],
                xyline[i_interval + 1, 1] - xyline[i_interval, 1],
            )
        return easting, northing, angle

    def _compute_hwidth(self) -> None:
        """Compute half of the road width.

        Parameter is set at self.parms.hwidth, so nothing is returned.
        """
        # Compute half width of the road (i.e., distance from centerline to boundary of the road)
        self.parms.hwidth = self.options.lanewidth * self.options.nlanes / 2

    def __str__(self) -> str:
        """For debugging purposes, it can be nice to see which vertices are used for this road.

        :return: Short description of the way.
        """
        return f"Way[vertices={self.ivs}]"


def apply_offset(xy_data: np.ndarray, offset: Union[float, List[float]]) -> np.ndarray:
    """Apply an offset to a line defined by the xy-coordinates.

    :param xy_data: N-by-2 array with (x,y)-coordinates.
    :param offset: Single offset or list of offsets. In case of list of offsets, there are
                   two options.
                   - A list of 2 floats defining the offsets at both ends.
                     Using linear interpolation, the offsets at the other nodes are determined.
                   - A list of N floats. In this case, node i will have an offset of 'offset[i]'.
    :return: N-by-2 array with new (x,y)-coordinates.
    """
    # Determine the offsets, such that it is a list with N floats.
    if isinstance(offset, list):
        if len(offset) == 2:  # noqa: PLR2004
            distances = np.array(
                [
                    np.hypot(xy_data[i, 0] - xy_data[i - 1, 0], xy_data[i, 1] - xy_data[i - 1, 1])
                    for i in range(1, xy_data.shape[0])
                ]
            )
            cumulative_dist = np.concatenate(([0], np.cumsum(distances)))
            off = offset[0] + cumulative_dist / cumulative_dist[-1] * (offset[1] - offset[0])
        else:
            off = offset
    else:
        off = np.ones(xy_data.shape[0]) * offset

    direction = np.zeros_like(xy_data)
    xy_new = np.zeros_like(xy_data)
    for i in range(xy_data.shape[0]):
        direction[i, :] = xy_data[max(i, 1), :] - xy_data[max(i - 1, 0), :]
        direction[i, :] /= np.sqrt(np.dot(direction[i, :], direction[i, :]))

    # Compute the angle (alpha) between two different sections. We need the tan(alpha/2),
    # which is computed using the formula tan(alpha/2) = sqrt(tan_halpha_squared), with
    # tan_halpha_squared = (1-cos(alpha)) / (1+cos(alpha)).
    cosines = np.concatenate((np.sum(direction[:-1, :] * direction[1:, :], axis=1), [1]))
    tan_halpha_squared = (1 - cosines) / (1 + cosines)  # It is assumed that cosines is not -1 !
    tan_halpha_squared[tan_halpha_squared < 0] = 0
    sign = np.concatenate(
        (np.sign(direction[:-1, 1] * direction[1:, 0] - direction[:-1, 0] * direction[1:, 1]), [1])
    )
    extra = np.sqrt(tan_halpha_squared) * sign

    # Apply offset
    xy_new[:, 0] = xy_data[:, 0] + (direction[:, 0] * extra - direction[:, 1]) * off
    xy_new[:, 1] = xy_data[:, 1] + (direction[:, 1] * extra + direction[:, 0]) * off
    return xy_new



"""Class that does not allow for creating new attributes.

Author(s): Erwin de Gelder
"""

from typing import Any


class UnknownOptionError(Exception):
    """Error to be raised in case an unknown options would be used."""

    def __init__(self, key: str) -> None:
        """Description of error.

        :param key: The name of the options that is invalid.
        """
        super().__init__(f"Option '{key:s}' is not a valid option.")


class Options:
    """Contain options of objects.

    The goal of this class is to create objects that does not allow for creating
    new attributes. In this way, this cannot be done accidentily, as this can
    generally cause very weird errors.
    """

    __isfrozen = False

    def __init__(self, **kwargs: Any) -> None:  # noqa: ANN401  # Allow Any
        """Create class for dealing with options.

        :param kwargs: define all options that need to be set.
        """
        for key, value in kwargs.items():
            try:
                self.__getattribute__(key)
            except AttributeError as attr_err:
                raise UnknownOptionError(key) from attr_err
            self.__setattr__(key, value)

        # Make sure that no new attributes are created outside the __init__ function.
        self.freeze()

    def __setattr__(self, key: str, value: Any) -> None:  # noqa: ANN401  # Allow Any
        """Set the attribute of this object.

        :param key: Name of the attribute to be set.
        :param value: Value of the attribute to be set.
        """
        if self.__isfrozen and not hasattr(self, key):
            raise FrozenOptionsError(self)
        object.__setattr__(self, key, value)

    def freeze(self) -> None:
        """Freeze the attributes, meaning that it is not possible to add attributes."""
        self.__isfrozen = True

    def unfreeze(self) -> None:
        """Unfreeze the attributes, meaning that it is possible to add attributes."""
        self.__isfrozen = False


class FrozenOptionsError(Exception):
    """Error to be raised when an option is being added while Options is in frozen mode."""

    def __init__(self, options: Options) -> None:
        """Description of error.

        :param options: Options for which another options is being added.
        """
        super().__init__(f"{options} is a frozen class")



"""Scripts for testing letters.py.

Author(s): Erwin de Gelder
"""

from pathlib import Path

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pytest

from traffic_scene_renderer import Letters, LettersOptions
from traffic_scene_renderer.utilities import hsl2rgb

from .test_static_objects import save_fig

mpl.use("Agg")


def test_all_letters() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-8, 8)
    axes.set_ylim(-0.8, 0.8)
    Letters(axes, "ABCDEFGHIJKLMNOPQRSTUVWXYZ", LettersOptions(length=1, width=15))
    save_fig(fig, axes, Path("letters") / "alfabet.png", 16)


def test_not_implemented_letter() -> None:
    fig, axes = plt.subplots()
    try:
        Letters(axes, "#")
    except NotImplementedError:
        pass
    else:
        pytest.fail("NotImplementedError should have been raised.")


def test_colored_alfabet() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-8, 8)
    axes.set_ylim(-0.8, 0.8)
    letters = Letters(axes, "ABCDEFGHIJKLMNOPQRSTUVWXYZ", LettersOptions(length=1, width=15))
    letters.change_color(face_color=(1, 0.5, 0.5), edge_color=(0.8, 0, 0))
    save_fig(fig, axes, Path("letters") / "colored.png", 16)


def test_move_letters() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-3.5, 3.5)
    axes.set_ylim(-3.5, 3.5)
    alfabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    for i in range(13):
        letters = Letters(
            axes,
            alfabet[2 * i : 2 * (i + 1)],
            LettersOptions(
                length=1,
                width=1,
                face_color=hsl2rgb(i / 13, 1, 0.6),
                edge_color=hsl2rgb(i / 13, 1, 0.3),
            ),
        )
        letters.change_pos(
            2.7 * np.sin(2 * np.pi * i / 13), 2.7 * np.cos(2 * np.pi * i / 13), 2 * np.pi * i / 13
        )
    save_fig(fig, axes, Path("letters") / "moved.png", 12)



"""Bus object.

Author(s): Erwin de Gelder
"""

from typing import Optional, Tuple, Union

import numpy as np
from matplotlib.axes import Axes

from .path_follower import PathFollower
from .polygon import Polygon
from .utilities import hsl2rgb, rgb2hsl
from .vehicle import Vehicle, VehicleOptions


class BusOptions(VehicleOptions):
    """The default values of the options of a bus.

    The following list shows the options (within parentheses default values):
        length (0): The length of the bus (by default based on width and AR).
        width (2.0): The width of the bus.
        aspect_ratio (2.69): Aspect ratio of the bus' shape.
        line_width (1): The line width for plotting the bus.
        x_position_init (0): The initial horizontal position of the bus.
        y_position_init (0): The initial vertical position of the bus.
        angle_init (0): The initial angle of the bus.
        color ((0, 0.4375, 0.75)): Main color of the bus.
        color2 (None): Secondary color of the truck. By default, lighter version
            of color.
        luminance_diff (0.3): If color2 is not set, specify luminance difference
            between color2 and color.
        windowcolor ([1, 1, 1]): The color of the front window.
        edgecolor ((0, 0, 0)): In case of a non-transparent truck, the color of
            the lines is set using this option.
        layer (2): The layer in which the truck will be plotted.
    """

    length: float = 0
    width: float = 2.5
    aspect_ratio: float = 2.69
    window_color: Tuple[float, float, float] = (1, 1, 1)
    color2: Optional[Tuple[float, float, float]] = None
    luminance_diff: float = 0.3

    def __init__(self, **kwargs: Union[float, Tuple[float, float, float]]) -> None:
        """Create a container for all options for a bus.

        :param kwargs: Use kwargs to set any option other then the default option.
        """
        VehicleOptions.__init__(self, **kwargs)
        if self.length == 0:
            self.length = self.width * self.aspect_ratio


class Bus(Vehicle):
    """Plot a bus on the scene.

    Create a bus object using Bus(axes, options).

    Attributes:
        options (BusOptions): All options. For a detailed description, see above.
        is_braking (bool): Whether the car is braking or not.
        axes (Axes): The axes that is used for plotting.
        path_follower (PathFollower): Optional object, used when vehicle needs to follow a path.
    """

    def __init__(
        self,
        axes: Axes,
        options: Optional[BusOptions] = None,
        path_follower: Optional[PathFollower] = None,
    ) -> None:
        """Create a bus object.

        :param axes: The axes on which the car is supposed to be drawn.
        :param options: Various options for the car, see CarOptions.
        :param path_follower: Path that the car needs to follow (if any).
        """
        if options is None:
            options = BusOptions()
        Vehicle.__init__(self, axes, options, path_follower)
        self.options: BusOptions

    def determine_color2(self) -> Tuple[float, float, float]:
        """If defined, just return color2. If not, return lighter version of color.

        :return: RGB tuple of color.
        """
        if self.options.color2 is None:
            hue, saturation, luminance = rgb2hsl(*self.options.color)
            luminance = min(1.0, max(0.0, luminance + self.options.luminance_diff))
            return hsl2rgb(hue, saturation, luminance)
        return self.options.color2

    def plot_vehicle(self) -> None:
        """Plot the bus on the axes."""
        # Plot the bus.
        yoffset = 8 + 84 * self.options.aspect_ratio
        yscale = 168 * self.options.aspect_ratio
        xoffset = 106
        xscale = 168
        yrear = 8 + 168 * self.options.aspect_ratio

        # Fill with main color, mirror.
        xdata = np.array([76, 57, 36, 31, 25, 23, 22, 22, 23, 24, 27, 31, 35])
        ydata = np.array(
            [
                8,
                9,
                10,
                11,
                17,
                21,
                24,
                yrear - 14,
                yrear - 9,
                yrear - 7,
                yrear - 3,
                yrear - 1,
                yrear,
            ]
        )
        xdata_new = (np.concatenate((xdata, 2 * xoffset - np.flipud(xdata))) - xoffset) / xscale
        ydata_new = (np.concatenate((ydata, np.flipud(ydata))) - yoffset) / yscale
        self.fills += (
            Polygon(
                self.axes,
                xdata_new * self.options.width,
                -ydata_new * self.options.length,
                facecolor=self.options.color,
                edgecolor=self.options.edgecolor,
                zorder=self.options.layer,
            ),
        )

        # Fill with window color, mirror.
        xdata = np.array([52, 34, 31, 31, 34, 78])
        ydata = np.array([12, 13, 16, 24, 63, 62])
        xdata_new = (np.concatenate((xdata, 2 * xoffset - np.flipud(xdata))) - xoffset) / xscale
        ydata_new = (np.concatenate((ydata, np.flipud(ydata))) - yoffset) / yscale
        self.fills += (
            Polygon(
                self.axes,
                xdata_new * self.options.width,
                -ydata_new * self.options.length,
                facecolor=self.options.window_color,
                edgecolor=self.options.edgecolor,
                zorder=self.options.layer,
                fixed_color=True,
            ),
        )

        # Squares with lighter color.
        xdata = np.array([76, 136, 136, 76])
        for ysquare in (149, yrear - 113):
            ydata = np.array([30, 30, -30, -30]) + ysquare
            self.fills += (
                Polygon(
                    self.axes,
                    (xdata - xoffset) / xscale * self.options.width,
                    -(ydata - yoffset) / yscale * self.options.length,
                    facecolor=self.determine_color2(),
                    edgecolor=self.options.edgecolor,
                    zorder=self.options.layer,
                    fixed_color=True,
                ),
            )

        # Mirrors, on both sides
        xdata = np.array([22, 15, 15, 19, 19, 17, 11, 9, 9, 13, 13, 22])
        ydata = np.array([48, 41, 23, 23, 21, 20, 20, 21, 23, 23, 43, 52])
        for xdata2 in (xdata, 212 - xdata):
            self.fills += (
                Polygon(
                    self.axes,
                    (xdata2 - xoffset) / xscale * self.options.width,
                    -(ydata - yoffset) / yscale * self.options.length,
                    facecolor=self.options.edgecolor,
                    edgecolor=self.options.edgecolor,
                    zorder=self.options.layer,
                    fixed_color=True,
                ),
            )

        # Circles of ventilation
        for y_pos in (149, yrear - 113):
            for radius in (9, 20):
                xdata = radius * np.cos(np.linspace(0, 2 * np.pi, 50))
                ydata = y_pos + radius * np.sin(np.linspace(0, 2 * np.pi, 50))
                self.plots += (
                    self.axes.plot(
                        xdata / xscale * self.options.width,
                        -(ydata - yoffset) / yscale * self.options.length,
                        color=self.options.edgecolor,
                        linewidth=self.options.line_width,
                        zorder=self.options.layer,
                    )[0],
                )

        # Lines
        for x_pos, y_pos1, y_pos2 in zip(
            (52, 64, 76, 88, 100, 76, 88, 100),
            (119, 119, 191, 191, 191, yrear - 71, yrear - 71, yrear - 71),
            (
                yrear - 28,
                yrear - 28,
                yrear - 155,
                yrear - 155,
                yrear - 155,
                yrear - 28,
                yrear - 28,
                yrear - 28,
            ),
        ):
            for x_pos2 in (x_pos, 212 - x_pos):
                xdata = (np.ones(2) * x_pos2 - xoffset) / xscale * self.options.width
                ydata = -(np.array([y_pos1, y_pos2]) - yoffset) / yscale * self.options.length
                self.plots += (
                    self.axes.plot(
                        xdata,
                        ydata,
                        color=self.options.edgecolor,
                        linewidth=self.options.line_width,
                        zorder=self.options.layer,
                    )[0],
                )

    def change_color(
        self,
        face_color: Optional[Tuple[float, float, float]] = None,
        edge_color: Optional[Tuple[float, float, float]] = None,
    ) -> None:
        """Change the colors of the filled areas and the plotted lines.

        :param face_color: The new color of the filled area.
        :param edge_color: The new color of the edge of the filled areas and the lines.
        """
        Vehicle.change_color(self, face_color, edge_color)

        # Force face color change for two squares, which are the 3nd and 4th filled areas.
        # Note that these have a slight different color.
        if face_color is not None:
            self.options.color = face_color
            for fill in self.fills[2:4]:
                if isinstance(fill, Polygon):
                    fill.set_facecolor_forced(self.determine_color2())



"""Scripts for testing path_follower.py.

Author(s): Erwin de Gelder
"""

import numpy as np
import pytest

from traffic_scene_renderer import PathFollower
from traffic_scene_renderer.path_follower import PathFollowerLengthNotSetError

XDATA = np.array([0.0, 0.0, 5.0, 5.0, 5.1, 5.1, 5.2, 5.2, 20.0])
YDATA = np.array([0.0, 5.0, 5.0, 4.9, 4.9, 4.8, 4.8, 4.7, 4.7])


def test_path_follower_creation() -> None:
    PathFollower(XDATA, YDATA, 1)


def test_path_follower_start_at_1m() -> None:
    # Let's start with lambda=0.2, meaning at (0.0, 1.0)
    path_follower = PathFollower(XDATA, YDATA, 1, (0, 0.2))
    assert path_follower.get_rear_xy() == (0.0, 1.0)


def test_path_follower_location_front() -> None:
    path_follower = PathFollower(XDATA, YDATA, 2)
    assert path_follower.get_location_front() == (1, 0.4)  # 40% of first segment


def test_path_follower_front_xy() -> None:
    path_follower = PathFollower(XDATA, YDATA, 5)
    assert path_follower.get_front_xy() == (0.0, 5.0)


def test_path_follower_move_4m() -> None:
    path_follower = PathFollower(XDATA, YDATA, np.sqrt(2))
    path_follower.move_vehicle(4)
    assert path_follower.get_center_coordinates() == pytest.approx((0.5, 4.5, np.pi / 4))
    path_follower.move_vehicle(2)
    assert path_follower.get_rear_xy() == pytest.approx((1.0, 5.0))
    path_follower.move_vehicle(5)
    assert path_follower.get_rear_xy() == pytest.approx((5.7, 4.7))


def test_path_follower_length_not_set_error() -> None:
    path_follower = PathFollower(XDATA, YDATA)
    try:
        path_follower.get_front_xy()
    except PathFollowerLengthNotSetError:
        pass
    else:
        pytest.fail("PathFollowerLengthNotSetError should be raised.")



"""Scripts for testing truck.py.

Author(s): Erwin de Gelder
"""

from pathlib import Path

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from traffic_scene_renderer import Truck, TruckOptions
from traffic_scene_renderer.utilities import hsl2rgb

from .test_static_objects import save_fig

mpl.use("Agg")


def test_truck_creation() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-2, 2)
    axes.set_ylim(-4, 4)
    Truck(axes)
    save_fig(fig, axes, Path("truck") / "standard_truck.png", 5)


def test_truck_braking() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-2, 2)
    axes.set_ylim(-6, 4)
    truck = Truck(axes)
    truck.start_braking(3)
    save_fig(fig, axes, Path("truck") / "truck_braking.png", 5)


def test_truck_trailers() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-12, 4)
    for i in range(5):
        truck = Truck(axes, TruckOptions(trailer=True, l_trailer=4 + 1.5 * i))
        truck.change_color(face_color=hsl2rgb(i / 5, 0.8, 0.5))
        truck.change_pos(-7 + 3.5 * i, 0)
    save_fig(fig, axes, Path("truck") / "truck_trailers.png", 10)


def test_change_trailer_angle() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-10, 4)
    axes.set_ylim(-3, 5)
    truck = Truck(axes, TruckOptions(trailer=True))
    truck.change_pos(0, 0, np.pi / 2)
    truck.change_trailer_angle(np.pi / 12)
    save_fig(fig, axes, Path("truck") / "truck_trailer_angle.png", 14)



"""Scripts for testing bus.py.

Author(s): Erwin de Gelder
"""

from pathlib import Path

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from traffic_scene_renderer import Bus, BusOptions

from .test_static_objects import save_fig

mpl.use("Agg")


def test_bus_creation() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-2, 2)
    axes.set_ylim(-4, 4)
    Bus(axes, BusOptions(color2=(0.8, 0.8, 1)))
    save_fig(fig, axes, Path("bus") / "standard_bus.png", 2)


def test_bus_different_colors() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-26, 26)
    axes.set_ylim(-4, 4)
    bus = Bus(axes, BusOptions(color=(0.5, 0.5, 0.5)))
    bus.change_pos(-24, 0)
    bus = Bus(axes, BusOptions(color=(0.8, 0, 0)))
    bus.change_pos(-20, 0)
    bus = Bus(axes, BusOptions(color=(0.8, 0.4, 0)))
    bus.change_pos(-16, 0)
    bus = Bus(axes, BusOptions(color=(0.8, 0.8, 0)))
    bus.change_pos(-12, 0)
    bus = Bus(axes, BusOptions(color=(0.4, 0.8, 0)))
    bus.change_pos(-8, 0)
    bus = Bus(axes, BusOptions(color=(0, 0.8, 0)))
    bus.change_pos(-4, 0)
    Bus(axes, BusOptions(color=(0, 0.8, 0.4)))
    bus = Bus(axes, BusOptions(color=(0, 0.8, 0.8)))
    bus.change_pos(4, 0)
    bus = Bus(axes, BusOptions(color=(0, 0.4, 0.8)))
    bus.change_pos(8, 0)
    bus = Bus(axes, BusOptions(color=(0, 0, 0.8)))
    bus.change_pos(12, 0)
    bus = Bus(axes, BusOptions(color=(0.4, 0, 0.8)))
    bus.change_pos(16, 0)
    bus = Bus(axes, BusOptions(color=(0.8, 0, 0.8)))
    bus.change_pos(20, 0)
    bus = Bus(axes, BusOptions(color=(0.8, 0, 0.4)))
    bus.change_pos(24, 0)
    save_fig(fig, axes, Path("bus") / "different_colors.png", 26)


def test_bus_change_color_and_pos() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(0, 8)
    axes.set_ylim(0, 4)
    bus = Bus(axes)
    bus.change_color(face_color=(1, 0, 0))
    bus.change_pos(4, 2, np.pi / 2)
    save_fig(fig, axes, Path("bus") / "recolored_bus.png", 4)



"""Scripts for testing options.py.

Author(s): Erwin de Gelder
"""

import pytest

from traffic_scene_renderer import FrozenOptionsError, Options, UnknownOptionError


class OptionsTest(Options):
    """Class to test the functionality of Options."""

    test: bool = True


def test_normal_functionality_options() -> None:
    my_options = OptionsTest(test=False)

    # We can add an option when we unfreeze the object
    my_options.unfreeze()
    my_options.test = False


def test_unknown_option_error() -> None:
    # Create an Options object and try to add an option that is not valid.
    try:
        Options(test=True)
    except UnknownOptionError:
        pass
    else:
        pytest.fail("UnknownOptionError should be raised.")


def test_frozen_options_error() -> None:
    # Create an Options object and then try to add an option that is not valid.
    my_options = OptionsTest()
    try:
        my_options.test_me_too = True
    except FrozenOptionsError:
        pass
    else:
        pytest.fail("FrozenOptionsError should be raised.")



"""Constructing a connection, which is used to connect two ways.

Author(s): Erwin de Gelder
"""

from typing import List, Tuple

import numpy as np

from .crossing import Crossing
from .options import Options
from .vertex import Vertex
from .way import Way


class ConnectionParameters(Options):
    """All kinds of parameters of the connection are contained in this class."""

    hlength_merge: float = 5
    connection_at_start: Tuple[bool, bool] = (True, True)
    angle: float = 0
    distance: Tuple[float, float] = (0, 0)
    xoffset_left: float = 0
    yoffset_left: float = 0
    xoffset_right: float = 0
    yoffset_right: float = 0


class InvalidConnectionError(Exception):
    """Error in case an invalid connection is created.

    A connection is only valid if either the start or end of one way has the same vertex as the
    start or end of the other way.
    """

    def __init__(self) -> None:
        """Description of error."""
        super().__init__(
            "Invalid connection. Two ways do not share a vertex at their start or endpoints."
        )


class Connection:
    """Connection that is used to connect two ways with each other.

    Attributes:
        idx (int): The index of connection.
        way1 (Way): The first way of this connection.
        way2 (Way): The second way of this connection.
        parms (ConnectionParameters): All kinds of parameters that are used for processing.
        vertex (Vertex): The vertex that is shared among the two ways.
        crossing (Crossing): In case the connection is modelled as crossing.
    """

    def __init__(self, index: int, way1: Way, way2: Way) -> None:
        """Create a connection to connect two ways.

        :param index: Index of the connection.
        :way1: First way.
        :way2: Second way.
        """
        self.idx = index
        self.parms = ConnectionParameters()
        # Make sure that ways is ordered in terms of number of lanes
        if way1.options.nlanes > way2.options.nlanes:
            self.way1, self.way2 = way2, way1
        else:
            self.way1, self.way2 = way1, way2

        # Set connection reference in ways
        if self.way1.ivs[0] == self.way2.ivs[0]:
            self.way1.parms.connection.i_start = self.idx
            self.way2.parms.connection.i_start = self.idx
            self.parms.connection_at_start = (True, True)
        elif self.way1.ivs[-1] == self.way2.ivs[0]:
            self.way1.parms.connection.i_end = self.idx
            self.way2.parms.connection.i_start = self.idx
            self.parms.connection_at_start = (False, True)
        elif self.way1.ivs[0] == self.way2.ivs[-1]:
            self.way1.parms.connection.i_start = self.idx
            self.way2.parms.connection.i_end = self.idx
            self.parms.connection_at_start = (True, False)
        elif self.way1.ivs[-1] == self.way2.ivs[-1]:
            self.way1.parms.connection.i_end = self.idx
            self.way2.parms.connection.i_end = self.idx
            self.parms.connection_at_start = (False, False)
        else:
            raise InvalidConnectionError

        # Set the vertex
        if self.parms.connection_at_start[0]:
            self.vertex = self.way1.vertices[0]
        else:
            self.vertex = self.way1.vertices[-1]

        # If the angle is more than 45 degrees, than just treat connection as a crossing
        self.crossing = None
        dx1, dy1 = self.direction(0)
        dx2, dy2 = self.direction(1)
        self.parms.angle = np.arccos(
            (dx1 * dx2 + dy1 * dy2) / (np.hypot(dx1, dy1) * np.hypot(dx2, dy2))
        )
        if np.abs(self.parms.angle) <= 3 * np.pi / 4:
            self.crossing = Crossing(0, self.vertex, [self.way1, self.way2])

    def process(self) -> Tuple[List[Vertex], List[int]]:
        """Process the connection for rendering.

        :return: List of new vertices and list of indices of the crossings that need to be
                 reprocessed.
        """
        # If connection can be treated as a crossing, process the crossing and we are done
        if self.crossing is not None:
            self.crossing.process()
            return [], []

        self.check_for_offset()
        vertices, i_crossings = self.compute_distance_to_next_vertex()
        self.compute_offset()
        self.set_offset()

        return vertices, i_crossings

    def check_for_offset(self) -> None:
        """Check if we need to apply an offset to correctly connect the ways."""
        diff_nlanes = self.way2.options.nlanes - self.way1.options.nlanes
        if diff_nlanes == 1:
            # As this is checked when creating this object, we know that way2 has more lanes than
            # way1.
            offset = diff_nlanes / 2 * self.way2.options.lanewidth

            # If there is a lane going to the left, we assume that there is an extra lane to
            # go to the left. Otherwise, assume that lane needs to be added to the right.
            if self.way2.options.turnlanes is not None and (
                self.way2.options.turnlanes.startswith("left")
                or self.way2.options.turnlanes.endswith("left")
            ):
                offset *= -1
            if self.parms.connection_at_start[1]:
                self.way2.apply_offset([-offset, 0])
            else:
                self.way2.apply_offset([0, offset])

    def compute_single_distance(
        self, i_way: int, way: Way, vertices: List[Vertex], i_crossings: List[int]
    ) -> float:
        """Compute distance from a single way from starting vertex to the next one.

        :param i_way: Index of the way (0 or 1).
        :param way: The way for which to calculate the distance.
        :param vertices: List of new vertices.
        :param i_crossings: List if indices of the crossings that need to be reprocessed.
        :return: Distance from connection to the next vertex.
        """
        if self.parms.connection_at_start[i_way]:
            x1_way, y1_way = way.vertices[0].xcoordinate, way.vertices[0].ycoordinate
            x2_way, y2_way = way.vertices[1].xcoordinate, way.vertices[1].ycoordinate
        else:
            x1_way, y1_way = way.vertices[-1].xcoordinate, way.vertices[-1].ycoordinate
            x2_way, y2_way = way.vertices[-2].xcoordinate, way.vertices[-2].ycoordinate
        distance = np.hypot(x2_way - x1_way, y2_way - y1_way)
        if distance > self.parms.hlength_merge:
            xnew = self.parms.hlength_merge / distance * (x2_way - x1_way) + x1_way
            ynew = self.parms.hlength_merge / distance * (y2_way - y1_way) + y1_way
            vertices.append(Vertex(-1, xnew, ynew))
            index = 1 if self.parms.connection_at_start[i_way] else -1
            redo_start, redo_end = way.insert_vertex(vertices[-1], index)
            if redo_start:
                i_crossings.append(way.parms.crossing.i_start)
            if redo_end:
                i_crossings.append(way.parms.crossing.i_end)
            distance = self.parms.hlength_merge
        return distance

    def compute_distance_to_next_vertex(self) -> Tuple[List[Vertex], List[int]]:
        """Compute the distance from the vertex of the connection to the next vertex of the ways.

        :return: List of new vertices and list of indices of the crossings that need to be
                 reprocessed.
        """
        vertices = []  # type: List[Vertex]
        i_crossings = []  # type: List[int]
        self.parms.distance = (
            self.compute_single_distance(0, self.way1, vertices, i_crossings),
            self.compute_single_distance(1, self.way2, vertices, i_crossings),
        )
        return vertices, i_crossings

    def compute_offset(self) -> None:
        """Compute the outer points of the ways at the connection."""
        xoffsets, yoffsets = np.zeros((2, 2)), np.zeros((2, 2))
        for i_way, way in enumerate([self.way1, self.way2]):
            # Find outer points of both ways
            if self.parms.connection_at_start[i_way]:
                x1_way, y1_way = way.vertices[0].xcoordinate, way.vertices[0].ycoordinate
            else:
                x1_way, y1_way = way.vertices[-1].xcoordinate, way.vertices[-1].ycoordinate
            x_difference, y_difference = self.direction(i_way)
            x_normal = -y_difference / np.hypot(x_difference, y_difference)
            y_normal = x_difference / np.hypot(x_difference, y_difference)
            if self.parms.connection_at_start[i_way]:
                offset = way.parms.offset[0]
            else:
                offset = -way.parms.offset[-1]
            xoffsets[0, i_way] = x1_way + x_normal * (offset + way.parms.hwidth)
            yoffsets[0, i_way] = y1_way + y_normal * (offset + way.parms.hwidth)
            xoffsets[1, i_way] = x1_way + x_normal * (offset - way.parms.hwidth)
            yoffsets[1, i_way] = y1_way + y_normal * (offset - way.parms.hwidth)

        # Compute weighted average of end points.
        self.parms.xoffset_left = (
            xoffsets[0, 0] * self.parms.distance[1] + xoffsets[1, 1] * self.parms.distance[0]
        ) / np.sum(self.parms.distance)
        self.parms.xoffset_right = (
            xoffsets[1, 0] * self.parms.distance[1] + xoffsets[0, 1] * self.parms.distance[0]
        ) / np.sum(self.parms.distance)
        self.parms.yoffset_left = (
            yoffsets[0, 0] * self.parms.distance[1] + yoffsets[1, 1] * self.parms.distance[0]
        ) / np.sum(self.parms.distance)
        self.parms.yoffset_right = (
            yoffsets[1, 0] * self.parms.distance[1] + yoffsets[0, 1] * self.parms.distance[0]
        ) / np.sum(self.parms.distance)

    def set_offset(self) -> None:
        """Set the offset as parameters for the ways, such that they are plotted correctly."""
        if self.parms.connection_at_start[0]:
            self.way1.parms.connection.x_left_start = self.parms.xoffset_left
            self.way1.parms.connection.x_right_start = self.parms.xoffset_right
            self.way1.parms.connection.y_left_start = self.parms.yoffset_left
            self.way1.parms.connection.y_right_start = self.parms.yoffset_right
        else:
            self.way1.parms.connection.x_left_end = self.parms.xoffset_right
            self.way1.parms.connection.x_right_end = self.parms.xoffset_left
            self.way1.parms.connection.y_left_end = self.parms.yoffset_right
            self.way1.parms.connection.y_right_end = self.parms.yoffset_left
        if self.parms.connection_at_start[1]:
            self.way2.parms.connection.x_left_start = self.parms.xoffset_right
            self.way2.parms.connection.x_right_start = self.parms.xoffset_left
            self.way2.parms.connection.y_left_start = self.parms.yoffset_right
            self.way2.parms.connection.y_right_start = self.parms.yoffset_left
        else:
            self.way2.parms.connection.x_left_end = self.parms.xoffset_left
            self.way2.parms.connection.x_right_end = self.parms.xoffset_right
            self.way2.parms.connection.y_left_end = self.parms.yoffset_left
            self.way2.parms.connection.y_right_end = self.parms.yoffset_right

    def direction(self, i: int) -> Tuple[float, float]:
        """Compute the direction (i.e., (dx, dy)) of the way.

        :param i: index of way, either 0 or 1.
        :return: (x, y)-direction.
        """
        way = self.way1 if i == 0 else self.way2
        if self.parms.connection_at_start[i]:
            x_dir = way.vertices[1].xcoordinate - way.vertices[0].xcoordinate
            y_dir = way.vertices[1].ycoordinate - way.vertices[0].ycoordinate
        else:
            x_dir = way.vertices[-2].xcoordinate - way.vertices[-1].xcoordinate
            y_dir = way.vertices[-2].ycoordinate - way.vertices[-1].ycoordinate
        return x_dir, y_dir



"""Traffic light.

Author(s): Erwin de Gelder
"""

from enum import Enum
from typing import Optional, Tuple

import numpy as np
from matplotlib.axes import Axes

from .options import Options
from .static_objects import StaticObject


class TrafficLightStatus(Enum):
    """Possible status for traffic light."""

    REMOVED = 0
    IDLE = 1
    RED = 2
    AMBER = 3
    GREEN = 4


class TrafficLightOptions(Options):
    """Class containing the default values of the options of a traffic light object."""

    radius: float = 0
    width: float = 0
    length: float = 0
    inter_dist: float = 0
    rectangle_color: Tuple[float, float, float] = (0, 0, 0)
    red_color: Tuple[float, float, float] = (1, 0, 0)
    amber_color: Tuple[float, float, float] = (1, 1, 0)
    green_color: Tuple[float, float, float] = (0, 1, 0)
    red_idle_color: Tuple[float, float, float] = (0.4, 0, 0)
    amber_idle_color: Tuple[float, float, float] = (0.4, 0.4, 0)
    green_idle_color: Tuple[float, float, float] = (0, 0.4, 0)
    signal_lines: int = 8
    signal_inner_radius: float = 0
    signal_outer_radius: float = 0
    amber: bool = True


class NoAmberError(Exception):
    """Error to be raised in case amber light is set while traffic light does not have amber."""

    def __init__(self) -> None:
        """Description of error."""
        super().__init__("Cannot show amber signal: traffic light signal has no amber.")


class TrafficLight(StaticObject):
    """A traffic light.

    Attributes:
        options (TrafficLightOptions): Options that can be set for the traffic light.
        axes (Axes): The axes that are used for plotting.
        status (TrafficLightStatus): The status of the traffic light.
    """

    def __init__(self, axes: Axes, options: Optional[TrafficLightOptions] = None) -> None:
        """Create a traffic light object.

        :param axes: The axes on which the traffic light is supposed to be drawn.
        :param options: Any options for the traffic light, see TrafficLightOptions.
        """
        self.options = TrafficLightOptions() if options is None else options
        self.axes = axes
        StaticObject.__init__(self, axes)
        if self.options.radius == 0:
            self.options.radius = np.diff(axes.get_xlim())[0] * 0.005
        if self.options.width == 0:
            self.options.width = self.options.radius * 2.6
        if self.options.length == 0:
            if self.options.amber:
                self.options.length = self.options.radius * 7.2
            else:
                self.options.length = self.options.radius * 4.9
        if self.options.inter_dist == 0:
            self.options.inter_dist = self.options.radius * 2.3
        if self.options.signal_inner_radius == 0:
            self.options.signal_inner_radius = self.options.radius * 1.8
        if self.options.signal_outer_radius == 0:
            self.options.signal_outer_radius = self.options.radius * 2.5
        self.status = TrafficLightStatus.REMOVED
        self.plot_idle()

    def plot_idle(self) -> None:
        """Plot an idle traffic light."""
        if self.status != TrafficLightStatus.REMOVED:
            self.remove()

        self.status = TrafficLightStatus.IDLE
        # Draw the rectangle.
        x_data = np.array(
            [
                -self.options.width / 2,
                self.options.width / 2,
                self.options.width / 2,
                -self.options.width / 2,
            ]
        )
        y_data = np.array(
            [
                self.options.length / 2,
                self.options.length / 2,
                -self.options.length / 2,
                -self.options.length / 2,
            ]
        )
        self.fills = (self.axes.fill(x_data, y_data, color=self.options.rectangle_color)[0],)

        # Draw the red, amber, and green signals.
        theta = np.linspace(0, 2 * np.pi, 30)
        x_data = self.options.radius * np.cos(theta)
        y_data = self.options.radius * np.sin(theta)
        if self.options.amber:
            self.fills += (
                self.axes.fill(
                    x_data, y_data + self.options.inter_dist, color=self.options.red_color
                )[0],
                self.axes.fill(x_data, y_data, color=self.options.amber_color)[0],
                self.axes.fill(
                    x_data, y_data - self.options.inter_dist, color=self.options.green_color
                )[0],
            )
        else:
            self.fills += (
                self.axes.fill(
                    x_data, y_data + 0.5 * self.options.inter_dist, color=self.options.red_color
                )[0],
                self.axes.fill(
                    x_data, y_data - 0.5 * self.options.inter_dist, color=self.options.green_color
                )[0],
            )

    def idle(self) -> None:
        """Plot an idle traffic light and reposition it."""
        self.plot_idle()
        self.set_position()

    def set_position(self) -> None:
        """Set the position of the traffic light correctly.

        It can happen that the position of the traffic light does not coincide with
        the information given by self.position. This happens when the traffic light
        is redrawn, e.g., using self.idle(). To set the position correcly again,
        this function can be used.
        """
        x_center, y_center = self.position.x_center, self.position.y_center
        angle = self.position.angle
        self.position.x_center, self.position.y_center, self.position.angle = 0, 0, 0
        self.change_pos(x_center, y_center, angle)

    def remove(self) -> None:
        """Remove the traffic light."""
        for fill in self.fills:
            fill.remove()
            del fill
        for plot in self.plots:
            plot.remove()
            del plot
        self.plots = ()
        self.fills = ()
        self.status = TrafficLightStatus.REMOVED

    def signal_data(self, signal: TrafficLightStatus) -> Tuple[np.ndarray, np.ndarray]:
        """Get the x and y data of the signal.

        :param signal: The signal to show.
        :return: x and y data of the signal.
        """
        xdata = np.zeros((2, self.options.signal_lines))
        ydata = np.array(np.zeros_like(xdata))
        for i, theta in enumerate(
            np.linspace(0, np.pi * 2, self.options.signal_lines, endpoint=False)
        ):
            xdata[0, i] = np.cos(theta) * self.options.signal_inner_radius
            xdata[1, i] = np.cos(theta) * self.options.signal_outer_radius
            ydata[0, i] = np.sin(theta) * self.options.signal_inner_radius
            ydata[1, i] = np.sin(theta) * self.options.signal_outer_radius
        if signal in (TrafficLightStatus.RED, TrafficLightStatus.GREEN):
            distance = self.options.inter_dist
            if not self.options.amber:
                distance = 0.5 * self.options.inter_dist
            if signal == TrafficLightStatus.RED:
                ydata += distance
            elif signal == TrafficLightStatus.GREEN:
                ydata -= distance
        return xdata, ydata

    def red(self) -> None:
        """Show red light signal."""
        self.remove()
        self.plot_idle()
        self.status = TrafficLightStatus.RED
        xdata, ydata = self.signal_data(self.status)
        self.plots = (self.axes.plot(xdata, ydata, color=self.options.red_color)[0],)
        if self.options.amber:
            self.fills[2].set_color(self.options.amber_idle_color)
            self.fills[3].set_color(self.options.green_idle_color)
        else:
            self.fills[2].set_color(self.options.green_idle_color)
        self.set_position()

    def amber(self) -> None:
        """Show amber light signal."""
        if not self.options.amber:
            raise NoAmberError
        self.remove()
        self.plot_idle()
        self.status = TrafficLightStatus.AMBER
        xdata, ydata = self.signal_data(self.status)
        self.plots = (self.axes.plot(xdata, ydata, color=self.options.amber_color)[0],)
        self.fills[1].set_color(self.options.red_idle_color)
        self.fills[3].set_color(self.options.green_idle_color)
        self.set_position()

    def green(self) -> None:
        """Show green light signal."""
        self.remove()
        self.plot_idle()
        self.status = TrafficLightStatus.GREEN
        xdata, ydata = self.signal_data(self.status)
        self.plots = (self.axes.plot(xdata, ydata, color=self.options.green_color)[0],)
        self.fills[1].set_color(self.options.red_idle_color)
        if self.options.amber:
            self.fills[2].set_color(self.options.amber_idle_color)
        self.set_position()

    def set_status(self, status: TrafficLightStatus) -> None:
        """Set the status of the traffic light.

        :param status: The new status of the traffic light.
        """
        if status == TrafficLightStatus.REMOVED:
            self.remove()
        elif status == TrafficLightStatus.IDLE:
            self.idle()
        elif status == TrafficLightStatus.RED:
            self.red()
        elif status == TrafficLightStatus.AMBER:
            self.amber()
        elif status == TrafficLightStatus.GREEN:
            self.green()

    def __str__(self) -> str:
        """Provide simple description of traffic light.

        :return: Text with current status of traffic light.
        """
        return f"Traffic light, status={self.status.name:s}"
